# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for AQT examples."""

from collections.abc import Callable

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2.examples.cnn import serve_calibrated_ptq
from aqt.jax.v2.examples.cnn import serve_ptq
from aqt.jax.v2.examples.cnn import serve_unquantized
import jax


def _dummy_dataset(
    ds_size: int, image_rng: jax.Array, label_rng: jax.Array
) -> dict[str, jax.Array]:
  return {
      "image": jax.random.uniform(key=image_rng, shape=(ds_size, 28, 28, 1)),
      "label": jax.random.randint(
          key=label_rng, shape=(ds_size,), minval=0, maxval=10
      ),
  }


class CnnExamplesSmokeTest(parameterized.TestCase):

  @parameterized.parameters([
      serve_calibrated_ptq.run,
      serve_ptq.run,
      serve_unquantized.run,
  ])
  def test_run_function(
      self,
      test_func: Callable[[dict[str, jax.Array], dict[str, jax.Array]], None],
  ):
    rng = jax.random.key(0)
    rng, train_image_rng = jax.random.split(rng)
    rng, train_label_rng = jax.random.split(rng)
    rng, test_image_rng = jax.random.split(rng)
    _, test_label_rng = jax.random.split(rng)
    train_ds = _dummy_dataset(
        ds_size=100,
        image_rng=train_image_rng,
        label_rng=train_label_rng,
    )
    test_ds = _dummy_dataset(
        ds_size=100,
        image_rng=test_image_rng,
        label_rng=test_label_rng,
    )
    # This is not a behavioral test. It only makes sure the AQT API usages in
    # the example code are compatible with the latest code base by ensuring
    # they run without errors.
    test_func(train_ds, test_ds)


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Benchmark for the ImageNet example."""
import tempfile
import time

from absl import flags
from absl.testing import absltest
from absl.testing.flagsaver import flagsaver
from aqt.jax_legacy.jax.imagenet import train
from aqt.jax_legacy.jax.imagenet.configs.paper import resnet50_bfloat16
from flax.testing import Benchmark
import jax
import numpy as np


# Parse absl flags test_srcdir and test_tmpdir.
jax.config.parse_flags_with_absl()


FLAGS = flags.FLAGS


class ImagenetBenchmark(Benchmark):
  """Benchmarks for the ImageNet Flax example."""

  @flagsaver
  def test_8x_v100_half_precision(self):
    """Run ImageNet on 8x V100 GPUs in half precision for 2 epochs."""
    model_dir = tempfile.mkdtemp()

    config = resnet50_bfloat16.get_config()
    FLAGS.hparams_config_dict = config
    FLAGS.batch_size = 2048
    FLAGS.half_precision = True
    FLAGS.num_epochs = 2
    FLAGS.model_dir = model_dir

    start_time = time.time()
    train.main([])
    benchmark_time = time.time() - start_time
    summaries = self.read_summaries(model_dir)

    # Summaries contain all the information necessary for the regression
    # metrics.
    wall_time, _, eval_accuracy = zip(*summaries['eval_accuracy'])
    wall_time = np.array(wall_time)
    sec_per_epoch = np.mean(wall_time[1:] - wall_time[:-1])
    end_accuracy = eval_accuracy[-1]

    # Assertions are deferred until the test finishes, so the metrics are
    # always reported and benchmark success is determined based on *all*
    # assertions.
    self.assertBetween(sec_per_epoch, 210, 240)
    self.assertBetween(end_accuracy, 0.06, 0.09)

    # Use the reporting API to report single or multiple metrics/extras.
    self.report_wall_time(benchmark_time)
    self.report_metrics({'sec_per_epoch': sec_per_epoch,
                         'accuracy': end_accuracy})
    self.report_extras({
        'description':
            'Toy 8 x V100 test for ImageNet ResNet50.',
        'model_name':
            'resnet50'
    })


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for wmt_mlperf.models."""

from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax_legacy.jax import get_bounds
from aqt.jax_legacy.jax import hlo_utils
from aqt.jax_legacy.jax import primitives
from aqt.jax_legacy.jax import quant_config
from aqt.jax_legacy.jax import test_utils
from aqt.jax_legacy.jax.quantization import QuantType
from aqt.jax_legacy.jax.wmt_mlperf import models
from aqt.jax_legacy.jax.wmt_mlperf import training_hparams_generator_lib
import flax
import jax
import jax.numpy as jnp
import numpy as onp


class ModelsTest(parameterized.TestCase):

  def setUp(self):
    super(ModelsTest, self).setUp()
    self.input_shape = (1, 1)
    self.target_shape = (1, 1)
    self.inputs = jnp.ones(self.input_shape, dtype=jnp.float32)
    self.target = jnp.ones(self.target_shape, dtype=jnp.float32)
    self.key = jax.random.PRNGKey(0)
    self.transformer_small_kwargs = {
        'vocab_size': 1,
        'output_vocab_size': 1,
        'max_len': 1,
        'train': False,
    }
    self.transformer_full_kwargs = {
        'vocab_size': 4,
        'output_vocab_size': 4,
        'max_len': 2,
        'train': False,
    }

  def init_model(self, transformer_kwargs):
    model = models.Transformer(
        use_bfloat16=False,
        dynamic_context=quant_config.DynamicContext(
            collect_acts_stats=False, update_bounds=False),
        dropout_rate=.1,
        attention_dropout_rate=.1,
        should_decode=False,
        **transformer_kwargs)
    state = model.init(self.key, jnp.zeros(self.input_shape, jnp.float32),
                       jnp.zeros(self.target_shape, jnp.float32))
    return model, state

  @parameterized.named_parameters(
      dict(testcase_name='test_mlp_weight_quant_8bit', mlp_weight_prec=8),
      dict(testcase_name='test_mlp_weight_quant_4bit', mlp_weight_prec=4),
      dict(testcase_name='test_mlp_weight_quant_1bit', mlp_weight_prec=1),
  )
  @mock.patch.object(primitives, 'round_with_gradient')
  @mock.patch.object(primitives, 'floor_with_gradient')
  def test_mlp_weight_quant(self, floor_with_gradient, round_with_gradient,
                            mlp_weight_prec):
    hparams = training_hparams_generator_lib.create_base_transformer_hparams(
        mlp_weight_prec=mlp_weight_prec,
        embedding_weight_prec=None,
        attention_weight_prec=None,
        mlp_pos_inputs_prec=None,
        mlp_pos_inputs_hyper=None,
        mlp_signed_inputs_prec=None,
        mlp_signed_inputs_hyper=None,
        attention_kqv_inputs_prec=None,
        attention_kqv_inputs_hyper=None,
        attention_out_inputs_prec=None,
        attention_out_inputs_hyper=None,
        logits_inputs_prec=None,
        logits_inputs_hyper=None,
        logits_via_embeddings=True,
        attention_act_q_inputs_prec=None,
        attention_act_q_inputs_hyper=None,
        attention_act_k_inputs_prec=None,
        attention_act_k_inputs_hyper=None,
        attention_act_probs_inputs_prec=None,
        attention_act_v_inputs_prec=None,
        attention_act_v_inputs_hyper=None,
        num_layers=1,
        emb_dim=1,
        num_heads=1,
        qkv_dim=1,
        mlp_dim=1,
        quant_type=QuantType.FAKE_QUANT)
    transformer_kwargs = self.transformer_small_kwargs
    transformer_kwargs['hparams'] = hparams
    round_with_gradient.side_effect = lambda x: x
    floor_with_gradient.side_effect = lambda x: x

    model, init_state = self.init_model(transformer_kwargs)
    # there are 2 MLP blocks in this model, with 2 quant ops each, so both clip
    # and round should be called 4 times each.
    round_with_gradient.assert_called_with(mock.ANY)
    self.assertEqual(round_with_gradient.call_count, 4)
    floor_with_gradient.assert_not_called()

    round_with_gradient.reset_mock()
    floor_with_gradient.reset_mock()

    output = model.apply(init_state, self.inputs, self.target)
    self.assertEqual(output.shape, (1, 1))
    round_with_gradient.assert_called_with(mock.ANY)
    self.assertEqual(round_with_gradient.call_count, 4)
    floor_with_gradient.assert_not_called()

  @mock.patch.object(primitives, 'round_with_gradient')
  @mock.patch.object(primitives, 'floor_with_gradient')
  def test_without_mlp_weight_quant(self, floor_with_gradient,
                                    round_with_gradient):
    hparams = training_hparams_generator_lib.create_base_transformer_hparams(
        mlp_weight_prec=None,
        embedding_weight_prec=None,
        attention_weight_prec=None,
        mlp_pos_inputs_prec=None,
        mlp_pos_inputs_hyper=None,
        mlp_signed_inputs_prec=None,
        mlp_signed_inputs_hyper=None,
        attention_kqv_inputs_prec=None,
        attention_kqv_inputs_hyper=None,
        attention_out_inputs_prec=None,
        attention_out_inputs_hyper=None,
        logits_inputs_prec=None,
        logits_inputs_hyper=None,
        logits_via_embeddings=True,
        attention_act_q_inputs_prec=None,
        attention_act_q_inputs_hyper=None,
        attention_act_k_inputs_prec=None,
        attention_act_k_inputs_hyper=None,
        attention_act_probs_inputs_prec=None,
        attention_act_v_inputs_prec=None,
        attention_act_v_inputs_hyper=None,
        num_layers=1,
        emb_dim=1,
        num_heads=1,
        qkv_dim=1,
        mlp_dim=1,
        quant_type=QuantType.FAKE_QUANT)
    transformer_kwargs = self.transformer_small_kwargs
    transformer_kwargs['hparams'] = hparams
    round_with_gradient.side_effect = lambda x: x
    floor_with_gradient.side_effect = lambda x: x

    model, init_state = self.init_model(transformer_kwargs)
    model.apply(init_state, self.inputs, self.target)
    round_with_gradient.assert_not_called()
    floor_with_gradient.assert_not_called()

  def _num_mlp_floors(self, weight_quant, pos_input_quant, neg_input_quant):
    # There are 2 MLP blocks per layer (1 encoder, 1 decoder) and 2 weight quant
    # ops per MLP block, so 4 in total per layer.
    mlp_floors_per_layer = 4 if weight_quant else 0

    # There are 2 MLP blocks per layer (1 encoder, 1 decoder) and 1 input quant
    # op per unsigned MLP block, so 2 in total per layer.
    if pos_input_quant:
      mlp_floors_per_layer = mlp_floors_per_layer + 2

    # There are 2 MLP blocks per layer (1 encoder, 1 decoder) and 1 input quant
    # op per signed MLP block, so 2 in total per layer.
    if neg_input_quant:
      mlp_floors_per_layer = mlp_floors_per_layer + 2

    return mlp_floors_per_layer

  def _num_embedding_floors(self, weight_quant, act_quant):
    # 3 embedding layers in the whole model
    embedding_floors = 3 if weight_quant else 0

    # logits activation quantization
    if act_quant:
      embedding_floors = embedding_floors + 1

    return embedding_floors

  def _num_attention_floors(self, weight_quant, kqv_input_quant,
                            out_input_quant, act_q_input_quant,
                            act_k_input_quant, act_qk_input_quant,
                            act_v_input_quant):
    # 3 attention blocks per layer (1 on encoder, 2 on decoder), each
    # attention block has 4 weight quant ops, so 12 in total per layer.
    attention_floors_per_layer = 12 if weight_quant else 0

    # 3 attention blocks per layer (1 on encoder, 2 on decoder), each
    # attention block has 3 kqv activation quant ops, so 9 in total per layer.
    if kqv_input_quant:
      attention_floors_per_layer = attention_floors_per_layer + 9

    # 3 attention blocks per layer (1 on encoder, 2 on decoder), each attention
    # block has 1 dense out activation quant op, so 3 in total per layer.
    if out_input_quant:
      attention_floors_per_layer = attention_floors_per_layer + 3

    # 3 attention blocks per layer (1 on encoder, 2 on decoder), each attention
    # block has 1 act*act activation quant op for each act per layer.
    for act_quant in [
        act_q_input_quant, act_k_input_quant, act_qk_input_quant,
        act_v_input_quant
    ]:
      if act_quant:
        attention_floors_per_layer = attention_floors_per_layer + 3

    return attention_floors_per_layer

  @parameterized.named_parameters(
      dict(
          testcase_name='test_2layers_no_quant',
          num_layers=2,
          mlp_weight_prec=None,
          mlp_pos_inputs_prec=None,
          mlp_signed_inputs_prec=None,
          attention_kqv_inputs_prec=None,
          attention_out_inputs_prec=None,
          embedding_weight_prec=None,
          attention_weight_prec=None,
          logits_inputs_prec=None,
          attention_act_q_inputs_prec=None,
          attention_act_k_inputs_prec=None,
          attention_act_probs_inputs_prec=None,
          attention_act_v_inputs_prec=None,
      ),
      dict(
          testcase_name='test_3layers_full_quant',
          num_layers=3,
          mlp_weight_prec=4,
          mlp_pos_inputs_prec=8,
          mlp_signed_inputs_prec=8,
          attention_kqv_inputs_prec=2,
          attention_out_inputs_prec=4,
          embedding_weight_prec=4,
          attention_weight_prec=4,
          logits_inputs_prec=8,
          attention_act_q_inputs_prec=4,
          attention_act_k_inputs_prec=8,
          attention_act_probs_inputs_prec=8,
          attention_act_v_inputs_prec=4,
      ),
  )
  def test_number_of_floor_ops(
      self, num_layers, mlp_weight_prec, mlp_pos_inputs_prec,
      mlp_signed_inputs_prec, attention_kqv_inputs_prec,
      attention_out_inputs_prec, embedding_weight_prec, attention_weight_prec,
      logits_inputs_prec, attention_act_q_inputs_prec,
      attention_act_k_inputs_prec, attention_act_probs_inputs_prec,
      attention_act_v_inputs_prec):
    # Counts number of floor ops as a proxy for quantization ops.
    act_fixed_clip_bound = 3.0
    hparams = training_hparams_generator_lib.create_base_transformer_hparams(
        mlp_weight_prec=mlp_weight_prec,
        embedding_weight_prec=embedding_weight_prec,
        attention_weight_prec=attention_weight_prec,
        mlp_pos_inputs_prec=mlp_pos_inputs_prec,
        mlp_pos_inputs_hyper=act_fixed_clip_bound,
        mlp_signed_inputs_prec=mlp_signed_inputs_prec,
        mlp_signed_inputs_hyper=act_fixed_clip_bound,
        attention_kqv_inputs_prec=attention_kqv_inputs_prec,
        attention_kqv_inputs_hyper=act_fixed_clip_bound,
        attention_out_inputs_prec=attention_out_inputs_prec,
        attention_out_inputs_hyper=act_fixed_clip_bound,
        logits_inputs_prec=logits_inputs_prec,
        logits_inputs_hyper=act_fixed_clip_bound,
        logits_via_embeddings=True,
        attention_act_q_inputs_prec=attention_act_q_inputs_prec,
        attention_act_q_inputs_hyper=act_fixed_clip_bound,
        attention_act_k_inputs_prec=attention_act_k_inputs_prec,
        attention_act_k_inputs_hyper=act_fixed_clip_bound,
        attention_act_probs_inputs_prec=attention_act_probs_inputs_prec,
        attention_act_v_inputs_prec=attention_act_v_inputs_prec,
        attention_act_v_inputs_hyper=act_fixed_clip_bound,
        num_layers=num_layers,
        emb_dim=5,
        num_heads=8,
        qkv_dim=8,
        mlp_dim=7,
        quant_type=QuantType.FAKE_QUANT)

    transformer_kwargs = self.transformer_full_kwargs
    transformer_kwargs['hparams'] = hparams
    input_shape = (2, 4)
    target_shape = input_shape
    model, init_state = self.init_model(transformer_kwargs)
    hlo_proto = hlo_utils.load_hlo_proto_from_model(model, init_state,
                                                    [input_shape, target_shape])
    floor_count = hlo_utils.count_ops_in_hlo_proto(hlo_proto, r'floor')

    mlp_floors_per_layer = self._num_mlp_floors(
        (mlp_weight_prec is not None), (mlp_pos_inputs_prec is not None),
        (mlp_signed_inputs_prec is not None))

    attention_floors_per_layer = self._num_attention_floors(
        (attention_weight_prec is not None),
        (attention_kqv_inputs_prec is not None),
        (attention_out_inputs_prec is not None),
        (attention_act_q_inputs_prec is not None),
        (attention_act_k_inputs_prec is not None),
        (attention_act_probs_inputs_prec is not None),
        (attention_act_v_inputs_prec is not None))

    embedding_floors = self._num_embedding_floors(
        (embedding_weight_prec is not None), (logits_inputs_prec is not None))

    expected_floor_count = num_layers * (
        mlp_floors_per_layer + attention_floors_per_layer) + embedding_floors
    self.assertEqual(floor_count, expected_floor_count)

  @parameterized.named_parameters(
      dict(
          testcase_name='test_2layers_att8bit_weight_quant',
          num_layers=2,
          attention_kqv_inputs_prec=None,
          attention_out_inputs_prec=None,
          attention_weight_prec=8,
          attention_act_q_inputs_prec=None,
          attention_act_k_inputs_prec=None,
          attention_act_probs_inputs_prec=None,
          attention_act_v_inputs_prec=None,
          inputs_hyper_is_float=True,
      ),
      dict(
          testcase_name='test_3layers_att8bit_kqv_inputs_quant',
          num_layers=3,
          attention_kqv_inputs_prec=8,
          attention_out_inputs_prec=None,
          attention_weight_prec=None,
          attention_act_q_inputs_prec=None,
          attention_act_k_inputs_prec=None,
          attention_act_probs_inputs_prec=None,
          attention_act_v_inputs_prec=None,
          inputs_hyper_is_float=True,
      ),
      dict(
          testcase_name='test_3layers_att8bit_act_q_inputs_quant',
          num_layers=3,
          attention_kqv_inputs_prec=None,
          attention_out_inputs_prec=None,
          attention_weight_prec=None,
          attention_act_q_inputs_prec=8,
          attention_act_k_inputs_prec=None,
          attention_act_probs_inputs_prec=None,
          attention_act_v_inputs_prec=None,
          inputs_hyper_is_float=True,
      ),
      dict(
          testcase_name='test_2layers_att8bit_act_k_inputs_quant',
          num_layers=2,
          attention_kqv_inputs_prec=None,
          attention_out_inputs_prec=None,
          attention_weight_prec=None,
          attention_act_q_inputs_prec=None,
          attention_act_k_inputs_prec=8,
          attention_act_probs_inputs_prec=None,
          attention_act_v_inputs_prec=None,
          inputs_hyper_is_float=True,
      ),
      dict(
          testcase_name='test_3layers_att8bit_act_qk_inputs_quant',
          num_layers=3,
          attention_kqv_inputs_prec=None,
          attention_out_inputs_prec=None,
          attention_weight_prec=None,
          attention_act_q_inputs_prec=None,
          attention_act_k_inputs_prec=None,
          attention_act_probs_inputs_prec=4,
          attention_act_v_inputs_prec=None,
          inputs_hyper_is_float=True,
      ),
      dict(
          testcase_name='test_2layers_att8bit_act_v_inputs_quant',
          num_layers=2,
          attention_kqv_inputs_prec=None,
          attention_out_inputs_prec=None,
          attention_weight_prec=None,
          attention_act_q_inputs_prec=None,
          attention_act_k_inputs_prec=None,
          attention_act_probs_inputs_prec=None,
          attention_act_v_inputs_prec=8,
          inputs_hyper_is_float=True,
      ),
      dict(
          testcase_name='test_3layers_att8bit_kqv_inputs_auto_quant',
          num_layers=3,
          attention_kqv_inputs_prec=8,
          attention_out_inputs_prec=None,
          attention_weight_prec=None,
          attention_act_q_inputs_prec=None,
          attention_act_k_inputs_prec=None,
          attention_act_probs_inputs_prec=None,
          attention_act_v_inputs_prec=None,
          inputs_hyper_is_float=False,
      ),
      dict(
          testcase_name='test_2layers_att8bit_out_inputs_quant',
          num_layers=2,
          attention_kqv_inputs_prec=None,
          attention_out_inputs_prec=8,
          attention_weight_prec=None,
          attention_act_q_inputs_prec=None,
          attention_act_k_inputs_prec=None,
          attention_act_probs_inputs_prec=None,
          attention_act_v_inputs_prec=None,
          inputs_hyper_is_float=True,
      ),
      dict(
          testcase_name='test_2layers_att8bit_out_inputs_auto_quant',
          num_layers=2,
          attention_kqv_inputs_prec=None,
          attention_out_inputs_prec=8,
          attention_weight_prec=None,
          attention_act_q_inputs_prec=None,
          attention_act_k_inputs_prec=None,
          attention_act_probs_inputs_prec=None,
          attention_act_v_inputs_prec=None,
          inputs_hyper_is_float=False,
      ),
      dict(
          testcase_name='test_2layers_att_weight_kqv_out_quant',
          num_layers=2,
          attention_kqv_inputs_prec=8,
          attention_out_inputs_prec=4,
          attention_weight_prec=2,
          attention_act_q_inputs_prec=None,
          attention_act_k_inputs_prec=None,
          attention_act_probs_inputs_prec=None,
          attention_act_v_inputs_prec=None,
          inputs_hyper_is_float=True,
      ),
      dict(
          testcase_name='test_2layers_att_weight_kqv_out_act_quant',
          num_layers=2,
          attention_kqv_inputs_prec=8,
          attention_out_inputs_prec=4,
          attention_weight_prec=2,
          attention_act_q_inputs_prec=4,
          attention_act_k_inputs_prec=8,
          attention_act_probs_inputs_prec=4,
          attention_act_v_inputs_prec=2,
          inputs_hyper_is_float=True,
      ),
      dict(
          testcase_name='test_2layers_att_weight_kqv_out_auto_quant',
          num_layers=2,
          attention_kqv_inputs_prec=8,
          attention_out_inputs_prec=4,
          attention_weight_prec=2,
          attention_act_q_inputs_prec=None,
          attention_act_k_inputs_prec=None,
          attention_act_probs_inputs_prec=None,
          attention_act_v_inputs_prec=None,
          inputs_hyper_is_float=False,
      ),
      dict(
          testcase_name='test_3layers_att_weight_kqv_out_act_auto_quant',
          num_layers=3,
          attention_kqv_inputs_prec=8,
          attention_out_inputs_prec=4,
          attention_weight_prec=2,
          attention_act_q_inputs_prec=8,
          attention_act_k_inputs_prec=4,
          attention_act_probs_inputs_prec=4,
          attention_act_v_inputs_prec=2,
          inputs_hyper_is_float=False,
      ),
  )
  def test_number_of_floor_ops_attention(
      self,
      num_layers,
      attention_kqv_inputs_prec,
      attention_out_inputs_prec,
      attention_weight_prec,
      attention_act_q_inputs_prec,
      attention_act_k_inputs_prec,
      attention_act_probs_inputs_prec,
      attention_act_v_inputs_prec,
      inputs_hyper_is_float,
  ):
    # Counts number of floor ops as a proxy for quantization ops.
    if inputs_hyper_is_float:
      inputs_hyper = 6.0
    else:
      inputs_hyper = get_bounds.GetBounds.Hyper(
          initial_bound=6.0,
          stddev_coeff=3.0,
          absdev_coeff=2.0,
          mix_coeff=0.5,
          granularity=quant_config.QuantGranularity.PER_TENSOR)
    hparams = training_hparams_generator_lib.create_base_transformer_hparams(
        mlp_weight_prec=None,
        embedding_weight_prec=None,
        attention_weight_prec=attention_weight_prec,
        mlp_pos_inputs_prec=None,
        mlp_pos_inputs_hyper=None,
        mlp_signed_inputs_prec=None,
        mlp_signed_inputs_hyper=None,
        attention_kqv_inputs_prec=attention_kqv_inputs_prec,
        attention_kqv_inputs_hyper=inputs_hyper,
        attention_out_inputs_prec=attention_out_inputs_prec,
        attention_out_inputs_hyper=inputs_hyper,
        logits_inputs_prec=None,
        logits_inputs_hyper=None,
        logits_via_embeddings=True,
        attention_act_q_inputs_prec=attention_act_q_inputs_prec,
        attention_act_q_inputs_hyper=inputs_hyper,
        attention_act_k_inputs_prec=attention_act_k_inputs_prec,
        attention_act_k_inputs_hyper=inputs_hyper,
        attention_act_probs_inputs_prec=attention_act_probs_inputs_prec,
        attention_act_v_inputs_prec=attention_act_v_inputs_prec,
        attention_act_v_inputs_hyper=inputs_hyper,
        num_layers=num_layers,
        emb_dim=5,
        num_heads=8,
        qkv_dim=8,
        mlp_dim=7,
        quant_type=QuantType.FAKE_QUANT)

    transformer_kwargs = self.transformer_full_kwargs
    transformer_kwargs['hparams'] = hparams
    input_shape = (2, 4)
    target_shape = input_shape
    model, init_state = self.init_model(transformer_kwargs)
    hlo_proto = hlo_utils.load_hlo_proto_from_model(model, init_state,
                                                    [input_shape, target_shape])
    floor_count = hlo_utils.count_ops_in_hlo_proto(hlo_proto, r'floor')

    attention_floors_per_layer = self._num_attention_floors(
        (attention_weight_prec is not None),
        (attention_kqv_inputs_prec is not None),
        (attention_out_inputs_prec is not None),
        (attention_act_q_inputs_prec is not None),
        (attention_act_k_inputs_prec is not None),
        (attention_act_probs_inputs_prec is not None),
        (attention_act_v_inputs_prec is not None))

    expected_floor_count = num_layers * attention_floors_per_layer
    self.assertEqual(floor_count, expected_floor_count)

  @parameterized.named_parameters(
      dict(
          testcase_name='test_3layers_mlp8bit_weight_quant',
          num_layers=3,
          mlp_weight_prec=8,
          mlp_pos_inputs_prec=None,
          mlp_pos_inputs_hyper_is_float=True,
          mlp_signed_inputs_prec=None,
          mlp_signed_inputs_hyper_is_float=True,
      ),
      dict(
          testcase_name='test_2layers_mlp8bit_pos_inputs_quant',
          num_layers=2,
          mlp_weight_prec=None,
          mlp_pos_inputs_prec=8,
          mlp_pos_inputs_hyper_is_float=True,
          mlp_signed_inputs_prec=None,
          mlp_signed_inputs_hyper_is_float=True,
      ),
      dict(
          testcase_name='test_2layers_mlp8bit_pos_inputs_auto_quant',
          num_layers=2,
          mlp_weight_prec=None,
          mlp_pos_inputs_prec=8,
          mlp_pos_inputs_hyper_is_float=False,
          mlp_signed_inputs_prec=None,
          mlp_signed_inputs_hyper_is_float=True,
      ),
      dict(
          testcase_name='test_2layers_mlp8bit_pos_inputs_weights_quant',
          num_layers=2,
          mlp_weight_prec=8,
          mlp_pos_inputs_prec=8,
          mlp_pos_inputs_hyper_is_float=True,
          mlp_signed_inputs_prec=None,
          mlp_signed_inputs_hyper_is_float=True,
      ),
      dict(
          testcase_name='test_2layers_mlp8bit_neg_inputs_quant',
          num_layers=2,
          mlp_weight_prec=None,
          mlp_pos_inputs_prec=None,
          mlp_pos_inputs_hyper_is_float=True,
          mlp_signed_inputs_prec=8,
          mlp_signed_inputs_hyper_is_float=True,
      ),
      dict(
          testcase_name='test_2layers_mlp8bit_neg_inputs_auto_quant',
          num_layers=2,
          mlp_weight_prec=None,
          mlp_pos_inputs_prec=None,
          mlp_pos_inputs_hyper_is_float=True,
          mlp_signed_inputs_prec=8,
          mlp_signed_inputs_hyper_is_float=False,
      ),
      dict(
          testcase_name='test_2layers_mlp8bit_all_inputs_weights_quant',
          num_layers=2,
          mlp_weight_prec=8,
          mlp_pos_inputs_prec=8,
          mlp_pos_inputs_hyper_is_float=True,
          mlp_signed_inputs_prec=8,
          mlp_signed_inputs_hyper_is_float=True,
      ),
      dict(
          testcase_name='test_2layers_mlp8bit_all_inputs_weights_auto_quant',
          num_layers=2,
          mlp_weight_prec=8,
          mlp_pos_inputs_prec=8,
          mlp_pos_inputs_hyper_is_float=False,
          mlp_signed_inputs_prec=8,
          mlp_signed_inputs_hyper_is_float=False,
      ),
  )
  def test_number_of_floor_ops_mlp(self, num_layers, mlp_weight_prec,
                                   mlp_pos_inputs_prec,
                                   mlp_pos_inputs_hyper_is_float,
                                   mlp_signed_inputs_prec,
                                   mlp_signed_inputs_hyper_is_float):
    # Counts number of floor ops as a proxy for quantization ops.
    if mlp_pos_inputs_hyper_is_float:
      mlp_pos_inputs_hyper = 6.0
    else:
      mlp_pos_inputs_hyper = get_bounds.GetBounds.Hyper(
          initial_bound=6.0,
          stddev_coeff=3.0,
          absdev_coeff=2.0,
          mix_coeff=0.5,
          granularity=quant_config.QuantGranularity.PER_TENSOR)
    if mlp_signed_inputs_hyper_is_float:
      mlp_pos_inputs_hyper = 6.0
    else:
      mlp_pos_inputs_hyper = get_bounds.GetBounds.Hyper(
          initial_bound=6.0,
          stddev_coeff=3.0,
          absdev_coeff=2.0,
          mix_coeff=0.5,
          granularity=quant_config.QuantGranularity.PER_TENSOR)
    hparams = training_hparams_generator_lib.create_base_transformer_hparams(
        mlp_weight_prec=mlp_weight_prec,
        embedding_weight_prec=None,
        attention_weight_prec=None,
        mlp_pos_inputs_prec=mlp_pos_inputs_prec,
        mlp_pos_inputs_hyper=mlp_pos_inputs_hyper,
        mlp_signed_inputs_prec=mlp_signed_inputs_prec,
        mlp_signed_inputs_hyper=mlp_pos_inputs_hyper,
        attention_kqv_inputs_prec=None,
        attention_kqv_inputs_hyper=None,
        attention_out_inputs_prec=None,
        attention_out_inputs_hyper=None,
        logits_inputs_prec=None,
        logits_inputs_hyper=None,
        logits_via_embeddings=True,
        attention_act_q_inputs_prec=None,
        attention_act_q_inputs_hyper=None,
        attention_act_k_inputs_prec=None,
        attention_act_k_inputs_hyper=None,
        attention_act_probs_inputs_prec=None,
        attention_act_v_inputs_prec=None,
        attention_act_v_inputs_hyper=None,
        num_layers=num_layers,
        emb_dim=5,
        num_heads=8,
        qkv_dim=8,
        mlp_dim=7,
        quant_type=QuantType.FAKE_QUANT)

    transformer_kwargs = self.transformer_full_kwargs
    transformer_kwargs['hparams'] = hparams
    input_shape = (2, 4)
    target_shape = input_shape
    model, init_state = self.init_model(transformer_kwargs)
    hlo_proto = hlo_utils.load_hlo_proto_from_model(model, init_state,
                                                    [input_shape, target_shape])
    floor_count = hlo_utils.count_ops_in_hlo_proto(hlo_proto, r'floor')

    mlp_floors_per_layer = self._num_mlp_floors(
        (mlp_weight_prec is not None), (mlp_pos_inputs_prec is not None),
        (mlp_signed_inputs_prec is not None))

    expected_floor_count = num_layers * mlp_floors_per_layer
    self.assertEqual(floor_count, expected_floor_count)

  @parameterized.named_parameters(
      dict(
          testcase_name='test_3layers_embedding8bit_weight_quant',
          num_layers=3,
          embedding_weight_prec=8,
          logits_inputs_prec=None,
          logits_inputs_hyper_is_float=True,
          logits_via_embeddings=True,
      ),
      dict(
          testcase_name='test_2layers_embedding8bit_inputs_auto_quant',
          num_layers=2,
          embedding_weight_prec=None,
          logits_inputs_prec=8,
          logits_inputs_hyper_is_float=False,
          logits_via_embeddings=True,
      ),
      dict(
          testcase_name='test_2layers_embedding8bit_inputs_weights_quant_fixed',
          num_layers=2,
          embedding_weight_prec=8,
          logits_inputs_prec=8,
          logits_inputs_hyper_is_float=True,
          logits_via_embeddings=True,
      ),
      dict(
          testcase_name='test_2layers_embedding8bit_inputs_weights_auto_quant',
          num_layers=2,
          embedding_weight_prec=8,
          logits_inputs_prec=8,
          logits_inputs_hyper_is_float=False,
          logits_via_embeddings=True,
      ),
      dict(
          testcase_name='test_2layers_embedding8bit_without_logit_sharing',
          num_layers=2,
          embedding_weight_prec=8,
          logits_inputs_prec=8,
          logits_inputs_hyper_is_float=False,
          logits_via_embeddings=False,
      ),
  )
  def test_number_of_floor_ops_embedding(self, num_layers,
                                         embedding_weight_prec,
                                         logits_inputs_prec,
                                         logits_inputs_hyper_is_float,
                                         logits_via_embeddings):
    # Counts number of floor ops as a proxy for quantization ops.
    if logits_inputs_hyper_is_float:
      logits_inputs_hyper = 6.0
    else:
      logits_inputs_hyper = get_bounds.GetBounds.Hyper(
          initial_bound=6.0,
          stddev_coeff=3.0,
          absdev_coeff=2.0,
          mix_coeff=0.5,
          granularity=quant_config.QuantGranularity.PER_TENSOR)

    hparams = training_hparams_generator_lib.create_base_transformer_hparams(
        mlp_weight_prec=None,
        embedding_weight_prec=embedding_weight_prec,
        attention_weight_prec=None,
        mlp_pos_inputs_prec=None,
        mlp_pos_inputs_hyper=None,
        mlp_signed_inputs_prec=None,
        mlp_signed_inputs_hyper=None,
        attention_kqv_inputs_prec=None,
        attention_kqv_inputs_hyper=None,
        attention_out_inputs_prec=None,
        attention_out_inputs_hyper=None,
        logits_inputs_prec=logits_inputs_prec,
        logits_inputs_hyper=logits_inputs_hyper,
        logits_via_embeddings=logits_via_embeddings,
        attention_act_q_inputs_prec=None,
        attention_act_q_inputs_hyper=None,
        attention_act_k_inputs_prec=None,
        attention_act_k_inputs_hyper=None,
        attention_act_probs_inputs_prec=None,
        attention_act_v_inputs_prec=None,
        attention_act_v_inputs_hyper=None,
        num_layers=num_layers,
        emb_dim=5,
        num_heads=8,
        qkv_dim=8,
        mlp_dim=7,
        quant_type=QuantType.FAKE_QUANT)

    transformer_kwargs = self.transformer_full_kwargs
    transformer_kwargs['hparams'] = hparams
    input_shape = (2, 4)
    target_shape = input_shape
    model, init_state = self.init_model(transformer_kwargs)
    hlo_proto = hlo_utils.load_hlo_proto_from_model(model, init_state,
                                                    [input_shape, target_shape])
    floor_count = hlo_utils.count_ops_in_hlo_proto(hlo_proto, r'floor')

    embedding_floor_ops = self._num_embedding_floors(
        (embedding_weight_prec is not None), (logits_inputs_prec is not None))

    self.assertEqual(floor_count, embedding_floor_ops)

  def test_padding_mask(self):
    # Fuzzing test to make sure activation statistics aren't affected by padding
    # tokens.
    #
    # This tests works by changing the embedding of the padding token (token
    # with id '0') and making sure all the stats stay the same.
    #
    # It also tests that the stats *do* change when the embedding of a
    # non-padding token changes.
    inputs_hyper = get_bounds.GetBounds.Hyper(
        initial_bound=6.0,
        stddev_coeff=3.0,
        absdev_coeff=2.0,
        mix_coeff=0.5,
        granularity=quant_config.QuantGranularity.PER_CHANNEL)
    # Set logits_via_embedding to false so that the embedding of the padding
    # token doesn't affect the logits calculation at the end of the decoder.
    hparams = training_hparams_generator_lib.create_base_transformer_hparams(
        mlp_weight_prec=8,
        embedding_weight_prec=None,
        attention_weight_prec=8,
        mlp_pos_inputs_prec=8,
        mlp_pos_inputs_hyper=inputs_hyper,
        mlp_signed_inputs_prec=8,
        mlp_signed_inputs_hyper=inputs_hyper,
        attention_kqv_inputs_prec=8,
        attention_kqv_inputs_hyper=inputs_hyper,
        attention_out_inputs_prec=8,
        attention_out_inputs_hyper=inputs_hyper,
        logits_inputs_prec=8,
        logits_inputs_hyper=inputs_hyper,
        logits_via_embeddings=False,
        attention_act_q_inputs_prec=8,
        attention_act_q_inputs_hyper=inputs_hyper,
        attention_act_k_inputs_prec=8,
        attention_act_k_inputs_hyper=inputs_hyper,
        attention_act_probs_inputs_prec=8,
        attention_act_v_inputs_prec=8,
        attention_act_v_inputs_hyper=inputs_hyper,
        num_layers=2,
        emb_dim=5,
        num_heads=2,
        qkv_dim=4,
        mlp_dim=4,
        quant_type=QuantType.FAKE_QUANT)
    module = models.Transformer(
        hparams=hparams,
        dynamic_context=quant_config.DynamicContext(
            update_bounds=True, collect_acts_stats=True),
        vocab_size=3,
        output_vocab_size=3,
        max_len=10,
        train=False,
        use_bfloat16=False,
        dropout_rate=.1,
        attention_dropout_rate=.1,
        should_decode=False)
    key = jax.random.PRNGKey(0)
    # Mark the first token of the target and last token of the inputs as padding
    # tokens.
    targets = onp.array([[0, 2]])
    inputs = onp.array([[1, 0]])
    initial_state = module.init(key, inputs=inputs, targets=targets)
    # Change the embedding of the padding token.
    initial_state = flax.core.unfreeze(initial_state)
    initial_state['params']['shared_embedding']['embedding'] = initial_state[
        'params']['shared_embedding']['embedding'].at[0, :].set(10.0)
    module.train = True
    _, state1 = module.apply(
        flax.core.freeze(initial_state),
        inputs=inputs,
        targets=targets,
        mutable=True,
        rngs={'dropout': key})
    initial_state['params']['shared_embedding']['embedding'] = initial_state[
        'params']['shared_embedding']['embedding'].at[0, :].set(20.0)
    _, state2 = module.apply(
        flax.core.freeze(initial_state),
        inputs=inputs,
        targets=targets,
        mutable=True,
        rngs={'dropout': key})
    # This tests the statistics in both the GetBounds and StatsTag modules.
    # test_utils.assert_stats_are_equal(state1, state2)

    # Now we repeat the test, but changing the embedding of a non-padding token
    # (token with ID 1 here). We expect to see the stats change.
    # print(initial_state)
    initial_state['params']['shared_embedding']['embedding'] = initial_state[
        'params']['shared_embedding']['embedding'].at[1, :].set(10.0)
    _, state1 = module.apply(
        flax.core.freeze(initial_state),
        inputs=inputs,
        targets=targets,
        mutable=True,
        rngs={'dropout': key})
    initial_state['params']['shared_embedding']['embedding'] = initial_state[
        'params']['shared_embedding']['embedding'].at[1, :].set(200.0)
    _, state2 = module.apply(
        flax.core.freeze(initial_state),
        inputs=inputs,
        targets=targets,
        mutable=True,
        rngs={'dropout': key})
    print(initial_state['get_bounds']['encoder']['encoderblock_0']
          ['enc_self_att']['K']['bounds'])
    print(state1['get_bounds']['encoder']['encoderblock_0']['enc_self_att']['K']
          ['bounds'])
    print(state2['get_bounds']['encoder']['encoderblock_0']['enc_self_att']['K']
          ['bounds'])
    print('')
    test_utils.assert_stats_are_unequal(state1, state2)

  def test_hparams_without_logits_when_logits_not_shared_raises_error(self):
    # Create hparams without logits hparams by passing in
    # logits_via_embeddings=True.
    inputs_hyper = get_bounds.GetBounds.Hyper(
        initial_bound=6.0,
        stddev_coeff=3.0,
        absdev_coeff=2.0,
        mix_coeff=0.5,
        granularity=quant_config.QuantGranularity.PER_CHANNEL)
    hparams = training_hparams_generator_lib.create_base_transformer_hparams(
        mlp_weight_prec=8,
        embedding_weight_prec=None,
        attention_weight_prec=8,
        mlp_pos_inputs_prec=8,
        mlp_pos_inputs_hyper=inputs_hyper,
        mlp_signed_inputs_prec=8,
        mlp_signed_inputs_hyper=inputs_hyper,
        attention_kqv_inputs_prec=8,
        attention_kqv_inputs_hyper=inputs_hyper,
        attention_out_inputs_prec=8,
        attention_out_inputs_hyper=inputs_hyper,
        logits_inputs_prec=8,
        logits_inputs_hyper=inputs_hyper,
        logits_via_embeddings=True,
        attention_act_q_inputs_prec=8,
        attention_act_q_inputs_hyper=inputs_hyper,
        attention_act_k_inputs_prec=8,
        attention_act_k_inputs_hyper=inputs_hyper,
        attention_act_probs_inputs_prec=8,
        attention_act_v_inputs_prec=8,
        attention_act_v_inputs_hyper=inputs_hyper,
        num_layers=2,
        emb_dim=5,
        num_heads=2,
        qkv_dim=4,
        mlp_dim=4,
        quant_type=QuantType.FAKE_QUANT)

    self.assertIsNone(hparams.decoder.logits)

    # Now set logits_via_embedding in the model hparams to False.
    hparams.logits_via_embedding = False
    module = models.Transformer(
        hparams=hparams,
        dynamic_context=quant_config.DynamicContext(
            update_bounds=True, collect_acts_stats=True),
        vocab_size=3,
        output_vocab_size=3,
        max_len=10,
        use_bfloat16=False,
        train=False,
        dropout_rate=.1,
        attention_dropout_rate=.1,
        should_decode=False)
    key = jax.random.PRNGKey(0)
    # Mark the first token of the target and last token of the inputs as padding
    # tokens.
    targets = onp.array([[0, 2]])
    inputs = onp.array([[1, 0]])
    # Because the model is not sharing logits with embeddings, but the logits
    # hparams are missing, it should raise an error.
    with self.assertRaises(ValueError):
      module.init(key, inputs=inputs, targets=targets)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A simple CNN model and MNIST dataset function."""

from typing import Any

from flax import linen as nn
import jax
import jax.numpy as jnp
import tensorflow_datasets as tfds

ModuleDef = Any


class CNN(nn.Module):
  """A simple CNN model."""

  bn_use_stats: bool
  dot_general_cls: ModuleDef = lambda: jax.lax.dot_general
  einsum_cls: ModuleDef = lambda: jnp.einsum

  @nn.compact
  def __call__(self, x):
    use_running_avg = not self.bn_use_stats
    x = nn.Conv(features=32, kernel_size=(3, 3))(x)
    x = nn.BatchNorm(use_running_average=use_running_avg, dtype=x.dtype)(x)
    x = nn.relu(x)
    x = nn.avg_pool(x, window_shape=(2, 2), strides=(2, 2))
    x = nn.Conv(features=64, kernel_size=(3, 3))(x)
    x = nn.BatchNorm(use_running_average=use_running_avg, dtype=x.dtype)(x)
    x = nn.relu(x)
    x = nn.avg_pool(x, window_shape=(2, 2), strides=(2, 2))
    x = x.reshape((x.shape[0], -1))  # flatten
    x = nn.Dense(features=256, dot_general_cls=self.dot_general_cls)(x)
    x = nn.relu(x)
    x = nn.Dense(features=10, dot_general_cls=self.dot_general_cls)(x)

    identity = jnp.identity(10, dtype=x.dtype)
    x = self.einsum_cls()('ab,bc->ac', x, identity)
    return x


def get_datasets() -> tuple[dict[str, jax.Array], dict[str, jax.Array]]:
  """Load MNIST train and test datasets into memory."""
  print('get_datasets started')
  ds_builder = tfds.builder('mnist')
  ds_builder.download_and_prepare()
  train_ds = tfds.as_numpy(ds_builder.as_dataset(split='train', batch_size=-1))
  test_ds = tfds.as_numpy(ds_builder.as_dataset(split='test', batch_size=-1))
  train_ds['image'] = jnp.float32(train_ds['image']) / 255.0
  test_ds['image'] = jnp.float32(test_ds['image']) / 255.0
  print('get_datasets DONE')
  return train_ds, test_ds



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for report_utils.py."""
import json
import os
import pathlib
import shutil

from absl.testing import absltest
from absl.testing import parameterized

from aqt.jax_legacy.utils import report_utils
import dacite

import numpy as onp

EventSeries = report_utils.EventSeries


class ReportUtilsTest(parameterized.TestCase):

  def setUp(self):
    super(ReportUtilsTest, self).setUp()
    self.event_series_a = EventSeries(
        name='test_series_a',
        steps=onp.array([0, 2, 4, 6, 8, 10]),
        values=onp.array([1, 2, 3, 4, 5, 6]))
    self.event_series_b = EventSeries(
        name='test_series_b',
        steps=onp.array([0, 2, 4, 6, 8, 10, 12]),
        values=onp.array([1, 2, 3, 0, 0, 3, 6]))
    self.event_series_with_nans = EventSeries(
        name='test_series_with_nans',
        steps=onp.array([0, 2, 4, 6, 8, 10]),
        values=onp.array([1, 2, 3, 4, onp.nan, onp.nan]))
    self.eval_freq = 2

    self.events_dict = {
        'a': self.event_series_a,
        'b': self.event_series_b,
        'with_nans': self.event_series_with_nans
    }

  def assertEventSeriesEqual(self, x: EventSeries, y: EventSeries):
    self.assertEqual(x.name, y.name)
    onp.testing.assert_array_almost_equal(x.steps, y.steps)
    onp.testing.assert_array_almost_equal(x.values, y.values)
    if x.wall_times is None:
      self.assertEqual(x.wall_times, y.wall_times)
    else:
      onp.testing.assert_array_almost_equal(x.wall_times, y.wall_times)

  @parameterized.named_parameters(
      dict(
          testcase_name='wdsz_1_st_0',
          window_size_in_steps=1,
          step=0,
          exp=1,
      ),
      dict(
          testcase_name='wdsz_2_st_1',
          window_size_in_steps=2,
          step=1,
          exp=(1 + 2) / 2,
      ),
      dict(
          testcase_name='wdsz_2_st_2',
          window_size_in_steps=2,
          step=2,
          exp=2,
      ),
      dict(
          testcase_name='wdsz_4_st_0',
          window_size_in_steps=4,
          step=0,
          exp=(1 + 2) / 2,
      ),
      dict(
          testcase_name='wdsz_4_st_2',
          window_size_in_steps=4,
          step=2,
          exp=(1 + 2 + 3) / 3,
      ),
      dict(
          testcase_name='wdsz_12_st_6',
          window_size_in_steps=12,
          step=6,
          exp=(1 + 2 + 3 + 4 + 5 + 6) / 6,
      ),
      dict(
          testcase_name='wdsz_12_st_8',
          window_size_in_steps=12,
          step=8,
          exp=(2 + 3 + 4 + 5 + 6) / 5,
      ),
  )
  def test_apply_smoothing_about_step_with_rectangular_kernel(
      self, window_size_in_steps, step, exp):
    smoothing_kernel = report_utils.SmoothingKernel.RECTANGULAR
    rectangular_kernel_fn = smoothing_kernel.get_func(
        window_size_in_steps=window_size_in_steps)
    res = report_utils.apply_smoothing_about_step(self.event_series_a, step,
                                                  rectangular_kernel_fn)
    self.assertAlmostEqual(res, exp)

  @parameterized.named_parameters(
      dict(
          testcase_name='wdsz_1_st_0',
          window_size_in_steps=1,
          step=0,
          exp=1,
      ),
      dict(
          testcase_name='wdsz_3_st_1',
          window_size_in_steps=3,
          step=1,
          exp=(1 + 2) / 2,
      ),
      dict(
          testcase_name='wdsz_2_st_2',
          window_size_in_steps=2,
          step=2,
          exp=2,
      ),
      dict(
          testcase_name='wdsz_4_st_0',
          window_size_in_steps=4,
          step=0,
          exp=1,
      ),
      dict(
          testcase_name='wdsz_4_st_2',
          window_size_in_steps=4,
          step=2,
          exp=2,
      ),
      dict(
          testcase_name='wdsz_12_st_6',
          window_size_in_steps=12,
          step=6,
          exp=(2 * 2 + 3 * 4 + 4 * 6 + 5 * 4 + 6 * 2) / 18,
      ),
      dict(
          testcase_name='wdsz_12_st_8',
          window_size_in_steps=12,
          step=8,
          exp=(3 * 2 + 4 * 4 + 5 * 6 + 6 * 4) / 16,
      ),
  )
  def test_apply_smoothing_about_step_with_triangular_kernel(
      self, window_size_in_steps, step, exp):
    smoothing_kernel = report_utils.SmoothingKernel.TRIANGULAR
    kernel_fn = smoothing_kernel.get_func(
        window_size_in_steps=window_size_in_steps)
    res = report_utils.apply_smoothing_about_step(self.event_series_a, step,
                                                  kernel_fn)
    self.assertAlmostEqual(res, exp)

  @parameterized.named_parameters(
      dict(
          testcase_name='wdsz_0',
          event_series_key='a',
          window_size_in_steps=0,
          step=0,
      ),
      dict(
          testcase_name='with_nans',
          event_series_key='with_nans',
          window_size_in_steps=5,
          step=0,
      ),
  )
  def test_apply_smoothing_about_step_raise_value_error(self, event_series_key,
                                                        window_size_in_steps,
                                                        step):
    smoothing_kernel = report_utils.SmoothingKernel.RECTANGULAR
    rectangular_kernel_fn = smoothing_kernel.get_func(
        window_size_in_steps=window_size_in_steps)
    with self.assertRaises(ValueError):
      report_utils.apply_smoothing_about_step(
          self.events_dict[event_series_key], step, rectangular_kernel_fn)

  @parameterized.named_parameters(
      dict(
          testcase_name='wdsz_2',
          window_size_in_steps=2,
          exp=EventSeries(
              name='test_series_a',
              steps=onp.array([0, 2, 4, 6, 8, 10]),
              values=onp.array([1, 2, 3, 4, 5, 6])),
      ),
      dict(
          testcase_name='wdsz_4',
          window_size_in_steps=4,
          exp=EventSeries(
              name='test_series_a',
              steps=onp.array([0, 2, 4, 6, 8, 10]),
              values=onp.array([(1 + 2) / 2, (1 + 2 + 3) / 3, (2 + 3 + 4) / 3,
                                (3 + 4 + 5) / 3, (4 + 5 + 6) / 3,
                                (5 + 6) / 2])),
      ),
      dict(
          testcase_name='wdsz_6',
          window_size_in_steps=6,
          exp=EventSeries(
              name='test_series_a',
              steps=onp.array([0, 2, 4, 6, 8, 10]),
              values=onp.array([(1 + 2) / 2, (1 + 2 + 3) / 3, (2 + 3 + 4) / 3,
                                (3 + 4 + 5) / 3, (4 + 5 + 6) / 3,
                                (5 + 6) / 2])),
      ),
      dict(
          testcase_name='wdsz_12',
          window_size_in_steps=12,
          exp=EventSeries(
              name='test_series_a',
              steps=onp.array([0, 2, 4, 6, 8, 10]),
              values=onp.array([(1 + 2 + 3 + 4) / 4, (1 + 2 + 3 + 4 + 5) / 5,
                                (1 + 2 + 3 + 4 + 5 + 6) / 6,
                                (1 + 2 + 3 + 4 + 5 + 6) / 6,
                                (2 + 3 + 4 + 5 + 6) / 5, (3 + 4 + 5 + 6) / 4])),
      ),
  )
  def test_apply_smoothing_with_rectangular_kernel(self, window_size_in_steps,
                                                   exp):
    smoothing_kernel = report_utils.SmoothingKernel.RECTANGULAR
    rectangular_kernel_fn = smoothing_kernel.get_func(
        window_size_in_steps=window_size_in_steps)
    res = report_utils.apply_smoothing(self.event_series_a,
                                       rectangular_kernel_fn)
    self.assertEventSeriesEqual(res, exp)

  @parameterized.named_parameters(
      dict(
          testcase_name='wdsz_2',
          window_size_in_steps=2,
          exp=EventSeries(
              name='test_series_a',
              steps=onp.array([0, 2, 4, 6, 8, 10]),
              values=onp.array([1, 2, 3, 4, 5, 6])),
      ),
      dict(
          testcase_name='wdsz_4',
          window_size_in_steps=4,
          exp=EventSeries(
              name='test_series_a',
              steps=onp.array([0, 2, 4, 6, 8, 10]),
              values=onp.array([1, 2, 3, 4, 5, 6])),
      ),
      dict(
          testcase_name='wdsz_6',
          window_size_in_steps=6,
          exp=EventSeries(
              name='test_series_a',
              steps=onp.array([0, 2, 4, 6, 8, 10]),
              values=onp.array([(1 * 3 + 2) / 4, (1 + 2 * 3 + 3) / 5,
                                (2 + 3 * 3 + 4) / 5, (3 + 4 * 3 + 5) / 5,
                                (4 + 5 * 3 + 6) / 5, (5 + 6 * 3) / 4])),
      ),
      dict(
          testcase_name='wdsz_12',
          window_size_in_steps=12,
          exp=EventSeries(
              name='test_series_a',
              steps=onp.array([0, 2, 4, 6, 8, 10]),
              values=onp.array([(1 * 6 + 2 * 4 + 3 * 2) / 12,
                                (1 * 4 + 2 * 6 + 3 * 4 + 4 * 2) / 16,
                                (1 * 2 + 2 * 4 + 3 * 6 + 4 * 4 + 5 * 2) / 18,
                                (2 * 2 + 3 * 4 + 4 * 6 + 5 * 4 + 6 * 2) / 18,
                                (3 * 2 + 4 * 4 + 5 * 6 + 6 * 4) / 16,
                                (4 * 2 + 5 * 4 + 6 * 6) / 12])),
      ),
  )
  def test_apply_smoothing_with_triangular_kernel(self, window_size_in_steps,
                                                  exp):
    smoothing_kernel = report_utils.SmoothingKernel.TRIANGULAR
    kernel_fn = smoothing_kernel.get_func(
        window_size_in_steps=window_size_in_steps)
    res = report_utils.apply_smoothing(self.event_series_a, kernel_fn)
    self.assertEventSeriesEqual(res, exp)

  @parameterized.named_parameters(
      dict(
          testcase_name='two_mins',
          event_series_key='b',
          early_stop_agg=report_utils.MinOrMax.MIN,
          start_step=0,
          exp=6),
      dict(
          testcase_name='min_start_step_4',
          event_series_key='b',
          early_stop_agg=report_utils.MinOrMax.MIN,
          start_step=4,
          exp=6),
      dict(
          testcase_name='min_start_step_10',
          event_series_key='b',
          early_stop_agg=report_utils.MinOrMax.MIN,
          start_step=10,
          exp=10),
      dict(
          testcase_name='max_start_step_0',
          event_series_key='b',
          early_stop_agg=report_utils.MinOrMax.MAX,
          start_step=0,
          exp=12),
      dict(
          testcase_name='with_nans_min_start_step_0',
          event_series_key='with_nans',
          early_stop_agg=report_utils.MinOrMax.MIN,
          start_step=0,
          exp=0),
      dict(
          testcase_name='with_nans_min_start_step_2',
          event_series_key='with_nans',
          early_stop_agg=report_utils.MinOrMax.MIN,
          start_step=2,
          exp=2),
      dict(
          testcase_name='with_nans_min_start_step_after_nans',
          event_series_key='with_nans',
          early_stop_agg=report_utils.MinOrMax.MIN,
          start_step=8,
          exp=8),
      dict(
          testcase_name='with_nans_max_start_step_0',
          event_series_key='with_nans',
          early_stop_agg=report_utils.MinOrMax.MAX,
          start_step=0,
          exp=6),
  )
  def test_find_early_stop_step(self, event_series_key, early_stop_agg,
                                start_step, exp):
    early_stop_func = early_stop_agg.get_func()
    res = report_utils.find_early_stop_step(
        self.events_dict[event_series_key],
        early_stop_func=early_stop_func,
        start_step=start_step)
    self.assertEqual(res, exp)

  def test_find_early_stop_step_raises_value_error_on_too_large_start_step(
      self):
    start_step = 20
    with self.assertRaisesRegex(
        ValueError, 'event_series does not have events after start_step.'):
      report_utils.find_early_stop_step(
          self.event_series_b,
          early_stop_func=onp.argmin,
          start_step=start_step)

  @parameterized.named_parameters(
      dict(
          testcase_name='min', early_stop_agg=report_utils.MinOrMax.MIN, exp=0),
      dict(
          testcase_name='max', early_stop_agg=report_utils.MinOrMax.MAX, exp=1),
  )
  def test_get_early_stop_func(self, early_stop_agg, exp):
    early_stop_func = early_stop_agg.get_func()
    res = early_stop_func(onp.array([1, 2]))
    self.assertEqual(res, exp)

  def test_get_early_stop_func_raises_error_on_unknown_agg(self):
    with self.assertRaises(AttributeError):
      early_stop_agg = 'bad input'
      early_stop_agg.get_func()

  def test_get_smoothing_kernel_func(self):
    smoothing_kernel = report_utils.SmoothingKernel.RECTANGULAR
    rect_kernel_func = smoothing_kernel.get_func(window_size_in_steps=3)
    self.assertEqual(rect_kernel_func(1), 1.)
    self.assertEqual(rect_kernel_func(2), 0.)
    self.assertEqual(rect_kernel_func(3), 0.)

  def test_get_smoothing_kernel_func_raises_error_on_unknown_agg(self):
    smoothing_kernel = 'bad input'
    with self.assertRaises(AttributeError):
      smoothing_kernel.get_func(window_size_in_steps=3)

  def test_get_smoothing_kernel_func_raises_error_on_rect_kernel_without_window_size(
      self):
    smoothing_kernel = report_utils.SmoothingKernel.RECTANGULAR
    with self.assertRaises(ValueError):
      smoothing_kernel.get_func(window_size_in_steps=None)

  @parameterized.named_parameters(
      dict(
          testcase_name='no_nans', event_series_key='a', start_step=0,
          exp=None),
      dict(
          testcase_name='with_nans',
          event_series_key='with_nans',
          start_step=0,
          exp=8),
      dict(
          testcase_name='with_nans_start_step_after_nan',
          event_series_key='with_nans',
          start_step=10,
          exp=10),
  )
  def test_check_for_nans(self, event_series_key, start_step, exp):
    first_nan_step = report_utils.check_for_nans(
        self.events_dict[event_series_key], start_step=start_step)
    self.assertEqual(first_nan_step, exp)

  @parameterized.named_parameters(
      dict(
          testcase_name='no_nans',
          all_events={
              'train': {
                  'acc':
                      EventSeries(
                          name='acc',
                          steps=onp.array([0, 2, 4, 6, 8]),
                          values=onp.array([0, 1, 2, 3, 4])),
              },
              'eval': {
                  'loss':
                      EventSeries(
                          name='loss',
                          steps=onp.array([0, 2, 4, 6, 8]),
                          values=onp.array([2, 4, 2, 3, 4])),
              }
          },
          exp=None),
      dict(
          testcase_name='with_nans',
          all_events={
              'train': {
                  'acc':
                      EventSeries(
                          name='acc',
                          steps=onp.array([0, 2, 4, 6, 8]),
                          values=onp.array([0, 1, 2, 3, 4])),
              },
              'eval': {
                  'loss':
                      EventSeries(
                          name='loss',
                          steps=onp.array([0, 2, 4, 6, 8]),
                          values=onp.array([2, 4, 2, onp.nan, onp.nan])),
              }
          },
          exp=6),
      dict(
          testcase_name='with_nan_in_multiple_series',
          all_events={
              'train': {
                  'acc':
                      EventSeries(
                          name='acc',
                          steps=onp.array([0, 2, 4, 6, 8]),
                          values=onp.array([0, 1, onp.nan, 3, 4])),
              },
              'eval': {
                  'loss':
                      EventSeries(
                          name='loss',
                          steps=onp.array([0, 2, 4, 6, 8]),
                          values=onp.array([2, 4, 2, onp.nan, onp.nan])),
              }
          },
          exp=4),
  )
  def test_check_all_events_for_nans(self, all_events, exp):
    first_nan_step = report_utils.check_all_events_for_nans(all_events)
    self.assertEqual(first_nan_step, exp)

  @parameterized.named_parameters(
      dict(
          testcase_name='no_nans_no_smoothing',
          all_events={
              'train': {
                  'acc':
                      EventSeries(
                          name='acc',
                          steps=onp.array([0, 2, 4, 6, 8]),
                          values=onp.array([0, 1, 2, 3, 4])),
              },
              'eval': {
                  'loss':
                      EventSeries(
                          name='loss',
                          steps=onp.array([0, 2, 4, 6, 8]),
                          values=onp.array([2, 4, 1, 4, 4])),
              }
          },
          early_stop_step=4,
          smoothing=False,
          exp_agg_metrics={
              'train': {
                  'acc': 2,
              },
              'eval': {
                  'loss': 1,
              },
          }),
      dict(
          testcase_name='no_nans_with_smoothing',
          all_events={
              'train': {
                  'acc':
                      EventSeries(
                          name='acc',
                          steps=onp.array([0, 2, 4, 6, 8]),
                          values=onp.array([0, 1, 2, 3, 4])),
              },
              'eval': {
                  'loss':
                      EventSeries(
                          name='loss',
                          steps=onp.array([0, 2, 4, 6, 8]),
                          values=onp.array([2, 4, 1, 4, 4])),
              }
          },
          early_stop_step=4,
          smoothing=True,
          exp_agg_metrics={
              'train': {
                  'acc': 2,
              },
              'eval': {
                  'loss': 3,
              },
          }),
      dict(
          testcase_name='with_nans_no_smoothing',
          all_events={
              'train': {
                  'acc':
                      EventSeries(
                          name='acc',
                          steps=onp.array([0, 2, 4, 6, 8]),
                          values=onp.array([0, 1, 2, 3, 4])),
              },
              'eval': {
                  'loss':
                      EventSeries(
                          name='loss',
                          steps=onp.array([0, 2, 4, 6, 8]),
                          values=onp.array([2, 4, 1, onp.nan, onp.nan])),
              }
          },
          early_stop_step=4,
          smoothing=False,
          exp_agg_metrics={
              'train': {
                  'acc': 2,
              },
              'eval': {
                  'loss': 1,
              },
          }),
  )
  def test_get_agg_metrics_at_step(self, all_events, early_stop_step, smoothing,
                                   exp_agg_metrics):
    if smoothing:
      smoothing_kernel_fn = report_utils.SmoothingKernel.RECTANGULAR.get_func(
          window_size_in_steps=5)
    else:
      smoothing_kernel_fn = None

    agg_metrics = report_utils.get_agg_metrics_at_step(all_events,
                                                       early_stop_step,
                                                       smoothing_kernel_fn)
    self.assertEqual(agg_metrics, exp_agg_metrics)

  @parameterized.named_parameters(
      dict(
          testcase_name='no_nans',
          all_events={
              'train': {
                  'acc':
                      EventSeries(
                          name='acc',
                          steps=onp.array([0, 2, 4, 6, 8]),
                          values=onp.array([0, 1, 2, 3, 4])),
              },
              'eval': {
                  'loss':
                      EventSeries(
                          name='loss',
                          steps=onp.array([0, 2, 4, 6, 8]),
                          values=onp.array([2, 4, 2, 1, 3])),
              }
          },
          exp_stop_step=6,
          exp_first_nan_step=None,
          exp_agg_metrics={
              'train': {
                  'acc': 3,
              },
              'eval': {
                  'loss': 2,
              },
          },
          exp_agg_metrics_unsmoothed={
              'train': {
                  'acc': 3,
              },
              'eval': {
                  'loss': 1,
              }
          }),
      dict(
          testcase_name='with_nans',
          all_events={
              'train': {
                  'acc':
                      EventSeries(
                          name='acc',
                          steps=onp.array([0, 2, 4, 6, 8]),
                          values=onp.array([0, 1, 2, 3, 4])),
              },
              'eval': {
                  'loss':
                      EventSeries(
                          name='loss',
                          steps=onp.array([0, 2, 4, 6, 8]),
                          values=onp.array([2, 4, 1, onp.nan, onp.nan])),
              }
          },
          exp_stop_step=4,
          exp_first_nan_step=6,
          exp_agg_metrics=None,
          exp_agg_metrics_unsmoothed={
              'train': {
                  'acc': 2,
              },
              'eval': {
                  'loss': 1,
              }
          }),
  )
  def test_compute_agg_metrics_from_events(self, all_events, exp_stop_step,
                                           exp_first_nan_step, exp_agg_metrics,
                                           exp_agg_metrics_unsmoothed):

    agg_metrics_unsmoothed, agg_metrics, stop_step, first_nan_step = report_utils.compute_agg_metrics_from_events(
        all_events=all_events,
        early_stop_component='eval',
        early_stop_attr='loss',
        early_stop_agg=report_utils.MinOrMax.MIN,
        smoothing_kernel=report_utils.SmoothingKernel.RECTANGULAR,
        window_size_in_steps=5,
        start_step=2,
    )

    self.assertEqual(stop_step, exp_stop_step)
    self.assertEqual(first_nan_step, exp_first_nan_step)
    self.assertEqual(agg_metrics, exp_agg_metrics)
    self.assertEqual(agg_metrics_unsmoothed, exp_agg_metrics_unsmoothed)

  @parameterized.named_parameters(
      dict(
          testcase_name='dir_with_fileext',
          model_dir='/a/b/c.json',
          exp='a--b--c'),
      dict(
          testcase_name='dir_without_fileext',
          model_dir='/a/b/c',
          exp='a--b--c'),
  )
  def test_report_path_from_model_dir(self, model_dir, exp):
    res = report_utils.report_path_from_model_dir(pathlib.Path(model_dir))
    self.assertEqual(res, pathlib.Path(exp))

  def test_save_report(self):
    temp_dir = '/tmp/test_save_report'
    report = report_utils.ExperimentReport(
        model_dir=temp_dir,
        metrics=None,
        unsmoothed_metrics=None,
        early_stop_step=1234,
        num_train_steps=2345,
        report_query_args={'test_key': 'test_val'})
    report_path = report_utils.save_report(report, temp_dir)

    with open(os.path.join(report_path,
                           report_utils.REPORT_FILENAME)) as report_file:
      report_dict = json.load(report_file)
      report_from_file = dacite.from_dict(
          data_class=report_utils.ExperimentReport, data=report_dict)
      self.assertEqual(report, report_from_file)

    shutil.rmtree(temp_dir)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test for AQT pallas."""

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import aqt_tensor
from aqt.jax.v2.pallas import pallas_call as aqt_pl
from aqt.jax.v2.pallas import quantizer
import jax
from jax.experimental import pallas as pl
from jax.experimental.pallas import tpu as pltpu
import jax.numpy as jnp


QTensor = aqt_tensor.QTensor


class AqtPallasTest(parameterized.TestCase):

  @parameterized.parameters(
      (
          (1024, 1024),
          (1,),
          (256, 256),
      ),
      (
          (1024, 1024),
          (0,),
          (256, 256),
      ),
      (
          (10, 512, 1024),
          (1,),
          (1, 256, 256),
      ),
      (
          (10, 512, 1024),
          (2,),
          (1, 256, 256),
      ),
  )
  def test_pallas_call_with_single_arg(
      self, tensor_shape, calibration_axes, block_shape
  ):
    key = jax.random.PRNGKey(0)
    x = jax.random.uniform(key, tensor_shape, minval=-3, maxval=3)
    block_spec = pl.BlockSpec(block_shape, lambda *args: args)

    qx = quantizer.quant(x, 8, calibration_axes=calibration_axes)

    def pallas_dequant(qx, block_spec):
      grid = [
          ndim // blk_ndim for ndim, blk_ndim in zip(tensor_shape, block_shape)
      ]

      def dequant_kernel(qx: QTensor, out_ref):
        qx.qvalue = qx.qvalue[...]
        qx.scale[0] = qx.scale[0][...]
        out_ref[...] = qx.dequant()

      dequant_out = aqt_pl.pallas_call(
          dequant_kernel,
          grid=tuple(grid),
          in_specs=[block_spec],
          out_specs=block_spec,
          out_shape=jax.ShapeDtypeStruct(shape=tensor_shape, dtype=jnp.float32),
          interpret=False,
      )(qx)
      return dequant_out

    # `pallas_dequant` dequantizes the each tile of qx whereas qx.dequant()
    # dequantizes the whole tensor. They should produces the same result.
    self.assertTrue((pallas_dequant(qx, block_spec) == qx.dequant()).all())

  @parameterized.parameters(
      (
          (10, 512, 1024),
          (1,),
          (1, 256, 256),
          3.3
      ),
      (
          (10, 512, 1024),
          (2,),
          (1, 256, 256),
          4.1
      ),
  )
  def test_pallas_call_prefetch_scalar(
      self, tensor_shape, calibration_axes, block_shape, multiple
  ):
    key = jax.random.PRNGKey(0)
    x = jax.random.uniform(key, tensor_shape, minval=-3, maxval=3)

    def index_map(i, j, k, _):
      return i, j, k

    block_spec = pl.BlockSpec(block_shape, index_map)

    qx = quantizer.quant(x, 8, calibration_axes=calibration_axes)

    # All values of multiple are the same.
    multiple = jnp.empty((128,)).at[:].set(multiple)

    def dequant_and_multiply(qx, multiple, block_spec):
      grid = [
          ndim // blk_ndim for ndim, blk_ndim in zip(tensor_shape, block_shape)
      ]

      def dequant_kernel(s, qx: QTensor, out_ref):
        qx.qvalue = qx.qvalue[...]
        qx.scale[0] = qx.scale[0][...]
        out_ref[...] = s[0] * qx.dequant()

      dequant_out = aqt_pl.pallas_call(
          dequant_kernel,
          grid_spec=pltpu.PrefetchScalarGridSpec(
              num_scalar_prefetch=1,
              grid=tuple(grid),
              in_specs=[block_spec],
              out_specs=block_spec,
          ),
          out_shape=jax.ShapeDtypeStruct(shape=tensor_shape, dtype=jnp.float32),
          interpret=False,
      )(multiple, qx)
      return dequant_out

    # `pallas_dequant` dequantizes the each tile of qx whereas qx.dequant()
    # dequantizes the whole tensor. They should produces the same result.
    self.assertTrue(
        (
            dequant_and_multiply(qx, multiple, block_spec)
            == (multiple[0] * qx.dequant())
        ).all()
    )


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for sparsity_core_depr."""

import dataclasses
import typing

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax_legacy.jax.flax import struct as flax_struct
from aqt.jax_legacy.jax.sparsity import sparsity_core_depr
from aqt.jax_legacy.jax.sparsity.sparsity_core_depr import SparseHParams
from aqt.jax_legacy.jax.sparsity.sparsity_core_depr import Sparsity
import flax
from flax import linen as nn
import jax
from jax import numpy as jnp
from jax import random
import numpy as np


dataclass = (
    flax_struct.dataclass if not typing.TYPE_CHECKING else dataclasses.dataclass
)


class SparsityTest(parameterized.TestCase):

  def init_model(
      self,
      update_mask,
      apply_mask,
      unstruct_sparsity,
      structure_decay=False,
      num_update_sparsity=0,
      mask_decay_weight=0.0,
      prune_rate=(2, 4),
      sparse_ste=False,
  ):
    rng = random.PRNGKey(0)
    self.inputs = jnp.array([[3, 4, 6], [1, 2, 1]])
    if unstruct_sparsity:
      sparsity_hparams = SparseHParams(
          type='UNSTRUCTURED',
          prune_rate=0.2,
          structure_decay=structure_decay,
          mask_decay_weight=mask_decay_weight,
          sparse_ste=sparse_ste,
      )
    else:
      sparsity_hparams = SparseHParams(
          type='STRUCTURED_NM',
          prune_rate=prune_rate,
          structure_decay=structure_decay,
          mask_decay_weight=mask_decay_weight,
          sparse_ste=sparse_ste,
      )
    sparsity_module = Sparsity(sparsity_hparams=sparsity_hparams)
    init_mask = sparsity_module.init(
        rng,
        self.inputs,
        update_mask=update_mask,
        apply_mask=apply_mask,
        num_update_sparsity=num_update_sparsity,
    )
    return sparsity_module, init_mask

  @parameterized.named_parameters(('unstruct', True), ('struct', False))
  def test_init(self, unstruct_sparsity):
    _, init_state = self.init_model(False, False, unstruct_sparsity)
    init_state_mask = init_state['sparsity']['mask']
    np.testing.assert_array_equal(
        init_state_mask, [[True, True, True], [True, True, True]]
    )

  @parameterized.named_parameters(
      dict(
          testcase_name='initial_structure_decay',
          num_update_sparsity=0,
          out=[[3, 4, 6, 8], [1, 2, 1, 4]],
          mask=[[True, True, True, True], [True, True, True, True]],
      ),
      dict(
          testcase_name='first_iteration_structure_decay',
          num_update_sparsity=1,
          out=[[0, 4, 6, 8], [0, 2, 1, 4]],
          mask=[[False, True, True, True], [False, True, True, True]],
      ),
      dict(
          testcase_name='second_iteration_structure_decay',
          num_update_sparsity=2,
          out=[[0, 0, 0, 8], [0, 0, 0, 4]],
          mask=[[False, False, False, True], [False, False, False, True]],
      ),
      dict(
          testcase_name='third_iteration_structure_decay',
          num_update_sparsity=3,
          out=[[0, 0, 0, 8], [0, 0, 0, 4]],
          mask=[[False, False, False, True], [False, False, False, True]],
      ),
  )
  def test_structure_decay(self, num_update_sparsity, out, mask):
    model, init_state = self.init_model(
        update_mask=False,
        apply_mask=False,
        unstruct_sparsity=False,
        structure_decay=True,
        prune_rate=(4, 4))
    # We need inputs that are divisible by four.
    inputs = jnp.array([[3, 4, 6, 8], [1, 2, 1, 4]])
    model_out, state_0 = model.apply(
        init_state,
        inputs,
        update_mask=True,
        apply_mask=True,
        num_update_sparsity=num_update_sparsity,
        mutable='sparsity')
    np.testing.assert_array_equal(model_out, out)
    np.testing.assert_array_equal(state_0['sparsity']['mask'], mask)

  @parameterized.named_parameters(
      dict(
          testcase_name='initial_mask_decay',
          num_update_sparsity=0,
          out=[[3, 4, 6, 8], [1, 2, 1, 4]],
          mask=[[False, False, True, True], [False, True, False, True]]),
      dict(
          testcase_name='first_iteration_mask_decay',
          num_update_sparsity=1,
          out=[[0.9 * 3, 0.9 * 4, 6, 8], [0.9 * 1, 2, 0.9 * 1, 4]],
          mask=[[False, False, True, True], [False, True, False, True]]),
      dict(
          testcase_name='second_iteration_mask_decay',
          num_update_sparsity=2,
          out=[[0.8 * 3, 0.8 * 4, 6, 8], [0.8 * 1, 2, 0.8 * 1, 4]],
          mask=[[False, False, True, True], [False, True, False, True]]),
      dict(
          testcase_name='third_iteration_mask_decay',
          num_update_sparsity=3,
          out=[[0.7 * 3, 0.7 * 4, 6, 8], [0.7 * 1, 2, 0.7 * 1, 4]],
          mask=[[False, False, True, True], [False, True, False, True]]),
      dict(
          testcase_name='ninth_iteration_mask_decay',
          num_update_sparsity=9,
          out=[[0.1 * 3, 0.1 * 4, 6, 8], [0.1 * 1, 2, 0.1 * 1, 4]],
          mask=[[False, False, True, True], [False, True, False, True]]),
      dict(
          testcase_name='tenth_iteration_mask_decay',
          num_update_sparsity=10,
          out=[[0.0 * 3, 0.0 * 4, 6, 8], [0.0 * 1, 2, 0.0 * 1, 4]],
          mask=[[False, False, True, True], [False, True, False, True]]),
  )
  def test_mask_decay(self, num_update_sparsity, out, mask):
    model, init_state = self.init_model(
        update_mask=False,
        apply_mask=False,
        unstruct_sparsity=False,
        structure_decay=False,
        mask_decay_weight=0.1,
        prune_rate=(2, 4))
    # We need inputs that are divisible by four.
    inputs = jnp.array([[3, 4, 6, 8], [1, 2, 1, 4]])
    model_out, state_0 = model.apply(
        init_state,
        inputs,
        update_mask=True,
        apply_mask=True,
        num_update_sparsity=num_update_sparsity,
        mutable='sparsity')
    np.testing.assert_allclose(model_out, out)
    np.testing.assert_array_equal(state_0['sparsity']['mask'], mask)

  # TODO(ayazdan): Add a more general test for other forms of sparsity.
  def test_sorting_network_sparse_2_4(self):
    key = random.PRNGKey(0)
    inputs = random.normal(key, (2, 5, 16))
    print(inputs)
    # n = 4
    # k = 2
    m = 4
    n = 2
    alpha = 1.0
    # JAX Top-K Implementation.
    xs = jnp.stack(
        jnp.split(inputs, int(jnp.shape(inputs)[-1] / m), axis=-1), axis=0)
    topk = jax.lax.top_k(xs, k=n)
    mask = jax.numpy.sum(
        jax.nn.one_hot(topk[1],
                       jnp.shape(xs)[-1]), axis=len(jnp.shape(inputs)))
    topk_filter = jnp.where(
        jnp.equal(mask, 0.), jnp.full(list(jnp.shape(mask)), alpha - 1), mask)
    filtered = jnp.multiply(xs, topk_filter)
    ys = jnp.concatenate(
        jnp.moveaxis(filtered, 0, 0), axis=len(jnp.shape(inputs)) - 1)
    print(ys)
    ys_hat = sparsity_core_depr.prune_2_4(inputs)
    print(ys_hat)
    np.testing.assert_array_equal(ys, ys_hat)

  @parameterized.named_parameters(
      dict(
          testcase_name='row_wise_pruning',
          order='R',
          exp_output=[[0, 0, 3, 4, 0, 0, 7, 8],
                      [0, 0, 11, 12, 0, 0, 15, 16],
                      [0, 0, 19, 20, 0, 0, 23, 24],
                      [0, 0, 27, 28, 0, 0, 31, 32]]),
      dict(
          testcase_name='column_wise_pruning',
          order='C',
          exp_output=[[0, 0, 0, 0, 0, 0, 0, 0],
                      [0, 0, 0, 0, 0, 0, 0, 0],
                      [17, 18, 19, 20, 21, 22, 23, 24],
                      [25, 26, 27, 28, 29, 30, 31, 32]]))
  def test_column_row_pruning(self, order, exp_output):
    inputs = jnp.reshape(jnp.arange(1, 33), (4, 8))
    output = sparsity_core_depr.prune_inputs_n_m(inputs, n=2, m=4, order=order)
    np.testing.assert_array_equal(output, exp_output)

  @parameterized.named_parameters(
      dict(
          testcase_name='column_wise_pruning',
          order='C',
          exp_output=[[[0, 0, 0, 0],
                       [0, 0, 0, 0],
                       [9, 10, 11, 12],
                       [13, 14, 15, 16]],
                      [[0, 0, 0, 0],
                       [0, 0, 0, 0],
                       [25, 26, 27, 28],
                       [29, 30, 31, 32]],
                      [[0, 0, 0, 0],
                       [0, 0, 0, 0],
                       [41, 42, 43, 44],
                       [45, 46, 47, 48]],
                      [[0, 0, 0, 0],
                       [0, 0, 0, 0],
                       [57, 58, 59, 60],
                       [61, 62, 63, 64]]]))
  def test_3d_column_pruning(self, order, exp_output):
    inputs = jnp.reshape(jnp.arange(1, 65), (4, 4, 4))
    output = sparsity_core_depr.prune_inputs_n_m(inputs, n=2, m=4, order=order)
    np.testing.assert_array_equal(output, exp_output)

  @parameterized.named_parameters(
      dict(
          testcase_name='offset_zero',
          offset=0,
          exp_output=[[0, 0, 3, 14, 0, 16, 0, 18, 0, 20],
                      [0, 12, 13, 0, 15, 0, 17, 0, 19, 0]]),
      dict(
          testcase_name='offset_positive',
          offset=5,
          exp_output=[[0, 0, 0, 14, 0, 16, 0, 18, 0, 20],
                      [11, 12, 13, 0, 15, 0, 17, 0, 19, 0]]))
  def test_n_m_row_wise_sparsity_with_offset(self, offset, exp_output):
    inputs = jnp.array([[1, 2, 3, 14, 5, 16, 7, 18, 9, 20],
                        [11, 12, 13, 4, 15, 6, 17, 8, 19, 10]])

    output = sparsity_core_depr.prune_inputs_n_m(
        inputs, n=2, m=4, offset=offset
    )
    np.testing.assert_array_equal(output, exp_output)

  @parameterized.named_parameters(
      dict(
          testcase_name='update_mask_apply_mask',
          update_mask=True,
          apply_mask=True),
      dict(
          testcase_name='no_update_mask_apply_mask',
          update_mask=False,
          apply_mask=True),
      dict(
          testcase_name='update_mask_no_apply_mask',
          update_mask=True,
          apply_mask=False),
      dict(
          testcase_name='no_update_mask_no_apply_mask',
          update_mask=False,
          apply_mask=False),
  )
  def test_sr_ste_fwd_pass(self, update_mask, apply_mask):
    rng = random.PRNGKey(0)
    sparsity_hparams = SparseHParams(
        type='STRUCTURED_NM',
        prune_rate=(2, 4),
        sparse_ste=True,
        structure_decay=False,
        mask_decay_weight=0.0,
    )
    inputs = jnp.array([[3, 4, 6, 8], [1, 2, 1, 4]])
    sparsity_module = Sparsity(sparsity_hparams=sparsity_hparams)
    init_state = sparsity_module.init(
        rng,
        inputs,
        update_mask=False,
        apply_mask=True,
        num_update_sparsity=0.0)
    # We need inputs that are divisible by four.
    inputs = jnp.array([[3, 4, 6, 8], [1, 2, 1, 4]])
    out, state_0 = sparsity_module.apply(
        init_state,
        inputs,
        update_mask=update_mask,
        apply_mask=apply_mask,
        mutable='sparsity')
    state_0_mask = state_0['sparsity']['mask']
    if not apply_mask or not update_mask:
      np.testing.assert_array_equal(out, inputs)
    else:
      np.testing.assert_array_equal(out, [[0, 0, 6, 8], [0, 2, 0, 4]])
    if update_mask:
      np.testing.assert_array_equal(
          state_0_mask,
          [[False, False, True, True], [False, True, False, True]])
    else:
      np.testing.assert_array_equal(
          state_0_mask, [[True, True, True, True], [True, True, True, True]])

    inputs2 = jnp.array([[2., 3., 1., 1.], [2., -6., 1., 5.]])
    out, state = sparsity_module.apply(
        state_0,
        inputs2,
        update_mask=update_mask,
        apply_mask=apply_mask,
        mutable='sparsity')
    state_mask = state['sparsity']['mask']
    if not apply_mask or not update_mask:
      np.testing.assert_array_equal(out, inputs2)
    else:
      np.testing.assert_array_equal(out, [[2., 3., 0., 0.], [0., -6., 0., 5.]])
    if update_mask:
      np.testing.assert_array_equal(
          state_mask, [[True, True, False, False], [False, True, False, True]])
    else:
      np.testing.assert_array_equal(
          state_mask, [[True, True, True, True], [True, True, True, True]])

  def test_sr_ste_bwd_pass(self):

    class SingleLayer(nn.Module):

      @dataclass
      class HParams:
        sparsity: SparseHParams

      hparams: HParams

      @nn.compact
      def __call__(
          self,
          inputs: jnp.ndarray,
          update_mask: bool,
          apply_mask: bool
      ) -> jnp.ndarray:
        kernel = self.param('kernel', nn.initializers.ones, (1, 4))
        kernel = Sparsity(
            sparsity_hparams=self.hparams.sparsity, name='weight_sparsity')(
                kernel,
                update_mask=update_mask,
                apply_mask=apply_mask,
                num_update_sparsity=0)
        return jnp.multiply(inputs, kernel)

    rng = random.PRNGKey(0)
    layer_kwargs = {}
    layer_kwargs['hparams'] = SingleLayer.HParams(
        sparsity=SparseHParams(
            type='STRUCTURED_NM',
            prune_rate=(2, 4),
            sparse_ste=True,
            structure_decay=False,
            mask_decay_weight=0.0,
        ))

    inputs = jnp.array([[2., 3., -5., 6.]])

    def loss_fn(params, state):
      del state
      model = SingleLayer(**layer_kwargs)
      y, updated_state = model.apply({
          'params': params,
      },
                                     inputs,
                                     update_mask=True,
                                     apply_mask=True,
                                     mutable=True)
      total_loss = jnp.sum(y)
      return total_loss, updated_state

    @jax.jit
    def update_params(params, grads):
      params = jax.tree_util.tree_map(lambda p, g: p - g, params, grads)
      return params

    module = SingleLayer(**layer_kwargs)
    variables = module.init(
        rng, jnp.zeros(inputs.shape), update_mask=False, apply_mask=False)
    state, params = flax.core.pop(variables, 'params')
    del variables
    for _ in range(10):
      # At each iteration, the pruned weights are multiplied with
      # ste_weight = 0.0002 and added to corresponding gradients.
      # In this simple example, gradients are simply the inputs to the network.
      (_, state), grads = jax.value_and_grad(
          loss_fn, has_aux=True, allow_int=True)(params, state)
      np.testing.assert_allclose(
          grads['kernel'], inputs + 0.0002 * jnp.multiply(
              ~state['sparsity']['weight_sparsity']['mask'], params['kernel']))
      params = update_params(params, grads)

  # TODO(shivaniagrawal): Add tests for struct sparsity as well.

  @parameterized.named_parameters(
      dict(
          testcase_name='update_mask_apply_mask',
          update_mask=True,
          apply_mask=True,
      ),
      dict(
          testcase_name='no_update_mask_apply_mask',
          update_mask=False,
          apply_mask=True),
      dict(
          testcase_name='update_mask_no_apply_mask',
          update_mask=True,
          apply_mask=False),
      dict(
          testcase_name='no_update_mask_no_apply_mask',
          update_mask=False,
          apply_mask=False),
  )
  def test_mask(self, update_mask, apply_mask):
    model, init_state = self.init_model(
        update_mask, False, unstruct_sparsity=True)
    out, state_0 = model.apply(
        init_state,
        self.inputs,
        update_mask=update_mask,
        apply_mask=apply_mask,
        mutable='sparsity')

    state_0_mask = state_0['sparsity']['mask']
    if not apply_mask or not update_mask:
      np.testing.assert_array_equal(out, self.inputs)
    else:
      np.testing.assert_array_equal(out, [[3, 4, 6], [0, 2, 1]])
    if update_mask:
      np.testing.assert_array_equal(state_0_mask,
                                    [[True, True, True], [False, True, True]])
    else:
      np.testing.assert_array_equal(state_0_mask,
                                    [[True, True, True], [True, True, True]])

    inputs2 = jnp.array([[2, 3, 1], [5, 4, 2]])
    out, state = model.apply(
        state_0,
        inputs2,
        update_mask=update_mask,
        apply_mask=apply_mask,
        mutable='sparsity')
    state_mask = state['sparsity']['mask']
    if not apply_mask or not update_mask:
      np.testing.assert_array_equal(out, inputs2)
    else:
      np.testing.assert_array_equal(out, [[2, 3, 0], [5, 4, 2]])
    if update_mask:
      np.testing.assert_array_equal(state_mask,
                                    [[True, True, False], [True, True, True]])
    else:
      np.testing.assert_array_equal(state_mask,
                                    [[True, True, True], [True, True, True]])


class PruningParamsTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(sparse_type='STRUCTURED_NM', prune_rate=0.1),
      dict(sparse_type='UNSTRUCTURED', prune_rate=(4, 1)))
  def test_invalid_params(self, sparse_type, prune_rate):
    with self.assertRaisesRegex(
        AssertionError, 'prune rate should be either None for no pruning'):
      sparsity_core_depr.SparseHParams(type=sparse_type, prune_rate=prune_rate)

  @parameterized.parameters(
      dict(
          sparse_type='STRUCTURED_NM',
          prune_rate=(4, 1),
          mask_decay_weight=-0.1),
      dict(sparse_type='UNSTRUCTURED', prune_rate=0.2, mask_decay_weight=-0.1))
  def test_invalid_mask_decay_weight(self, sparse_type, prune_rate,
                                     mask_decay_weight):
    with self.assertRaisesRegex(AssertionError,
                                '.* `mask_decay_weight` must be positive.'):
      sparsity_core_depr.SparseHParams(
          type=sparse_type,
          prune_rate=prune_rate,
          mask_decay_weight=mask_decay_weight,
      )

  @parameterized.parameters(
      dict(
          sparse_type='STRUCTURED_NM',
          prune_rate=(4, 1),
          sparse_ste=True,
          mask_decay_weight=0.1),
      dict(
          sparse_type='UNSTRUCTURED',
          prune_rate=0.2,
          sparse_ste=True,
          mask_decay_weight=0.1))
  def test_invalid_sparse_ste_with_non_zero_mask_decay_weight(
      self, sparse_type, prune_rate, sparse_ste, mask_decay_weight):
    with self.assertRaisesRegex(ValueError,
                                'SR-STE only works with non-decaying mask.'):
      sparsity_core_depr.SparseHParams(
          type=sparse_type,
          prune_rate=prune_rate,
          sparse_ste=sparse_ste,
          mask_decay_weight=mask_decay_weight,
      )

  @parameterized.parameters(
      dict(
          sparse_type='STRUCTURED_NM',
          prune_rate=(4, 1),
          sparse_ste=True,
          structure_decay=True),
      dict(
          sparse_type='UNSTRUCTURED',
          prune_rate=0.2,
          sparse_ste=True,
          structure_decay=True))
  def test_invalid_sparse_ste_with_structure_decay(self, sparse_type,
                                                   prune_rate, sparse_ste,
                                                   structure_decay):
    with self.assertRaisesRegex(
        ValueError,
        'SR-STE only works with non-decaying sparse structure.'):
      sparsity_core_depr.SparseHParams(
          type=sparse_type,
          prune_rate=prune_rate,
          sparse_ste=sparse_ste,
          structure_decay=structure_decay,
      )

  @parameterized.parameters(
      dict(sparse_type='UNSTRUCTURED', prune_rate=0.2, sparse_ste=True))
  def test_invalid_sparse_ste_with_unstructured_sparsity(
      self, sparse_type, prune_rate, sparse_ste):
    with self.assertRaisesRegex(
        ValueError, 'SR-STE only works with structured sparsity'
    ):
      sparsity_core_depr.SparseHParams(
          type=sparse_type, prune_rate=prune_rate, sparse_ste=sparse_ste
      )

  @parameterized.parameters(
      dict(
          sparse_type='STRUCTURED_NM',
          prune_rate=(4, 1),
          error_msg='must be lower than prune_rate'),
      dict(
          sparse_type='UNSTRUCTURED',
          prune_rate=2.5,
          error_msg='sparsity ratio can not be > 1, provided prune_rate'))
  def test_invalid_prune_rate(self, sparse_type, prune_rate, error_msg):
    sparsity_hparams = sparsity_core_depr.SparseHParams(
        type=sparse_type, prune_rate=prune_rate
    )

    inputs = jnp.arange(12)
    with self.assertRaisesRegex(AssertionError, error_msg):
      sparsity_core_depr.get_sparsity_mask(inputs, sparsity_hparams)


class PruningFunctionalityTest(parameterized.TestCase):

  def test_pruning_mask(self):
    # Total number of parameters = 20
    inputs = jnp.array(np.random.rand(10, 2))
    mask = sparsity_core_depr.get_pruning_unstruct_mask(inputs, prune_rate=0.1)
    self.assertEqual(jnp.sum(mask), 18)

  def test_smallest_largest_magnitude_mask(self):
    # Total number of parameters = 20
    inputs = jnp.array(np.arange(20))
    mask = sparsity_core_depr.get_pruning_unstruct_mask(
        inputs, smallest=True, prune_rate=0.1
    )
    self.assertFalse(mask[0])
    self.assertFalse(mask[1])

    mask = sparsity_core_depr.get_pruning_unstruct_mask(
        inputs, smallest=False, prune_rate=0.1
    )
    self.assertFalse(mask[-1])
    self.assertFalse(mask[-2], 0)

  def test_prune_inputs_unstruct(self):
    # Total number of parameters = 20
    inputs = jnp.array(np.random.rand(10, 2))
    prune_input = sparsity_core_depr.prune_inputs_unstruct(inputs)
    self.assertEqual(jnp.sum(prune_input == 0), 2)

  def test_smallest_largest_magnitude_prune_unstruct(self):
    # Total number of parameters = 20
    inputs = jnp.array(np.arange(20))
    prune_input = sparsity_core_depr.prune_inputs_unstruct(
        inputs, prune_rate=0.1
    )
    self.assertEqual(prune_input[0], 0)
    self.assertEqual(prune_input[1], 0)

    prune_input = sparsity_core_depr.prune_inputs_unstruct(
        inputs, prune_rate=0.1, smallest=False
    )
    self.assertEqual(prune_input[-1], 0)
    self.assertEqual(prune_input[-2], 0)

  def test_prune_inputs_n_m(self):
    inputs = jnp.array(np.random.rand(10, 2, 4))
    out = sparsity_core_depr.prune_inputs_n_m(inputs, n=1, m=4)
    self.assertEqual(out.shape[0], inputs.shape[0])
    self.assertEqual(out.shape[1], inputs.shape[1])
    self.assertEqual(out.shape[2], inputs.shape[2])
    # Only 20 non-zero elements must exist after pruning.
    self.assertEqual(out[out != 0].shape[0], 20)
    self.assertEqual(
        list(np.argmax(inputs, axis=2).flatten()),
        list(np.argmax(out != 0, axis=2).flatten()))

  def test_order_not_valid(self):
    inputs = jnp.array(np.random.rand(10, 2, 4))
    with self.assertRaises(ValueError):
      _ = sparsity_core_depr.prune_inputs_n_m(inputs, n=1, m=4, order='X')

  def test_n_m_pruning_mask(self):
    inputs = jnp.array(np.random.rand(10, 2, 4))
    mask = sparsity_core_depr.get_pruning_n_m_mask(inputs, n=1, m=4)
    self.assertEqual(
        list(np.argmax(inputs, axis=2).flatten()),
        list(np.argmax(mask == 1, axis=2).flatten()))


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Train utils for training a ResNet model on the ImageNet dataset.

The data is loaded using tensorflow_datasets.
The function create_model() can accept an hparam file as input, which will
  determine the model's size, training parameters, quantization precisions, etc.
"""

import dataclasses
import functools
from typing import Any, Mapping
from aqt.jax_legacy.jax.imagenet import input_pipeline
from aqt.jax_legacy.jax.imagenet import models
from aqt.jax_legacy.jax.imagenet import pokebnn

import flax
from flax import jax_utils
from flax import struct
from flax.training import common_utils
from flax.training import dynamic_scale as dynamic_scale_lib
import jax
from jax import lax
import jax.nn
import jax.numpy as jnp
import optax


def create_model(
    key,
    batch_size,
    image_size,
    model_dtype,
    hparams,
    train=True,
    is_teacher=True,  # create vanilla resnet by default
    **kwargs):
  """Creates the ResNet model using hparams."""
  input_shape = (batch_size, image_size, image_size, 3)
  if is_teacher:  # create teacher model
    model = models.create_resnet(hparams, model_dtype, train, **kwargs)
  else:  # create PokeBNN model
    model = pokebnn.create_pokebnn(hparams=hparams, train=train, **kwargs)
  init_state = model.init(key, jnp.zeros(input_shape, dtype=model_dtype))
  return model, init_state


def cross_entropy_loss(logits, labels):
  logits = jax.nn.log_softmax(logits)
  return -jnp.sum(
      common_utils.onehot(labels, num_classes=1000) * logits) / labels.size


def kl_div_loss(logits, teacher_logits):
  # Attention: logits are *not* log_softmaxed
  # Attention: teacher logits are probability distributions (softmaxed)
  log_prob_x = jax.nn.log_softmax(logits)  # default axis is -1
  # prob_y = jax.nn.softmax(teacher_logits)  # default axis is -1
  return -jnp.sum(teacher_logits * log_prob_x) / logits.shape[0]


def compute_metrics(logits, labels):
  loss = cross_entropy_loss(logits, labels)
  accuracy = jnp.mean(jnp.argmax(logits, -1) == labels)
  metrics = {
      'loss': loss,
      'accuracy': accuracy,
  }
  metrics = lax.pmean(metrics, axis_name='batch')
  return metrics


def train_step(model, state, batch, hparams, update_bounds, quantize_weights,
               learning_rate_fn, teacher):
  """Perform a single training step."""

  # TODO(yichi): update the quantize_weights flag in dynamic_context
  dynamic_context = dataclasses.replace(
      model.dynamic_context,
      update_bounds=update_bounds,
      quantize_weights=quantize_weights)

  model = dataclasses.replace(model, dynamic_context=dynamic_context)

  def loss_fn(params):
    """loss function used for training."""
    variables = {'params': params, **state.mutable_vars}
    logits, new_model_state = model.apply(
        variables, batch['image'], mutable=['batch_stats', 'get_bounds'])
    # TODO(yichi): use the checkpoint and the 8-bit model to compute logits
    teacher_logits = teacher['model'](teacher['variables'], batch['image'],
                                      batch['label'])
    # TODO(yichi): replace cross_entropy_loss with KL div loss
    loss = kl_div_loss(logits, teacher_logits)
    weight_penalty_params = jax.tree.leaves(params)
    weight_decay = hparams.weight_decay
    weight_l2 = sum(
        [jnp.sum(x**2) for x in weight_penalty_params if x.ndim > 1])
    weight_penalty = weight_decay * 0.5 * weight_l2
    loss = loss + weight_penalty
    return loss, (new_model_state, logits)

  step = state.step
  dynamic_scale = state.dynamic_scale
  lr = learning_rate_fn(step)

  if dynamic_scale:
    grad_fn = dynamic_scale.value_and_grad(
        loss_fn, has_aux=True, axis_name='batch')
    dynamic_scale, is_fin, aux, grad = grad_fn(state.params)
    # dynamic loss takes care of averaging gradients across replicas
  else:
    grad_fn = jax.value_and_grad(loss_fn, has_aux=True)
    aux, grad = grad_fn(state.params)
    # Re-use same axis_name as in the call to `pmap(...train_step...)` below.
    grad = lax.pmean(grad, axis_name='batch')
  new_model_state, logits = aux[1]
  metrics = compute_metrics(logits, batch['label'])
  metrics['learning_rate'] = lr

  updates, new_opt_state = state.tx.update(grad, state.opt_state, state.params)
  new_params = optax.apply_updates(state.params, updates)
  new_state = state.replace(
      step=step+1,
      params=new_params,
      mutable_vars=new_model_state,
      opt_state=new_opt_state,
      dynamic_scale=dynamic_scale)

  if dynamic_scale:
    # if is_fin == False the gradients contain Inf/NaNs and the old optimizer
    # state should be restored.
    new_state = new_state.replace(
        opt_state=jax.tree.map(
            functools.partial(jnp.where, is_fin),
            new_state.opt_state,
            state.opt_state),
        params=jax.tree.map(
            functools.partial(jnp.where, is_fin),
            new_state.params,
            state.params))
    metrics['scale'] = dynamic_scale.scale

  return new_state, metrics


# pylint: disable=missing-function-docstring
def eval_step(model, state, batch, quantize_weights):
  # TODO(yichi): update the quantize_weights flag in dynamic_context
  dynamic_context = dataclasses.replace(
      model.dynamic_context, quantize_weights=quantize_weights)
  model = dataclasses.replace(model, dynamic_context=dynamic_context)

  variables = {'params': state.params, **state.mutable_vars}
  model = dataclasses.replace(model, train=False)
  logits = model.apply(variables, batch['image'], mutable=False)
  return compute_metrics(logits, batch['label'])


def prepare_tf_data(xs):
  """Convert a input batch from tf Tensors to numpy arrays."""
  local_device_count = jax.local_device_count()

  def _prepare(x):
    # Use _numpy() for zero-copy conversion between TF and NumPy.
    x = x._numpy()  # pylint: disable=protected-access

    # reshape (host_batch_size, height, width, 3) to
    # (local_devices, device_batch_size, height, width, 3)
    return x.reshape((local_device_count, -1) + x.shape[1:])

  return jax.tree.map(_prepare, xs)


def create_input_iter(batch_size, data_dir, image_size, dtype, train, cache):
  """Creates input data iterator."""
  ds = input_pipeline.load_split(
      batch_size,
      data_dir=data_dir,
      image_size=image_size,
      dtype=dtype,
      train=train,
      cache=cache)
  it = map(prepare_tf_data, ds)
  it = jax_utils.prefetch_to_device(it, 2)
  return it


class TrainState(struct.PyTreeNode):  # pytype: disable=invalid-function-definition  # dataclass_transform
  """Train state keeping track of step, parameters, optimizer state, and etc."""
  step: int
  params: flax.core.FrozenDict[str, Any]
  mutable_vars: Mapping[str, Any]
  tx: optax.GradientTransformation = struct.field(pytree_node=False)
  opt_state: optax.OptState
  dynamic_scale: dynamic_scale_lib.DynamicScale

  @classmethod
  def create(cls, *, params, tx, **kwargs):
    """Creates a new instance with `step=0` and initialized `opt_state`."""
    opt_state = tx.init(params)
    return cls(
        step=0,
        params=params,
        tx=tx,
        opt_state=opt_state,
        **kwargs,
    )


def sync_batch_stats(state):
  """Sync the batch statistics across replicas."""
  avg = jax.pmap(lambda x: lax.pmean(x, 'x'), 'x')

  new_mutable_vars = state.mutable_vars.copy(
      {'batch_stats': avg(state.mutable_vars['batch_stats'])})
  return state.replace(mutable_vars=new_mutable_vars)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Input pipeline for WMT datasets."""

import functools
import os
import tempfile
import time
from typing import List, Optional
from absl import logging

import jax

import tensorflow.compat.v2 as tf
import tensorflow_datasets as tfds
import tensorflow_text as tftxt

from sentencepiece import SentencePieceTrainer

AUTOTUNE = tf.data.experimental.AUTOTUNE


# -----------------------------------------------------------------------------
# Raw TFDS dataset.
# -----------------------------------------------------------------------------
def raw_wmt_datasets(dataset_name='wmt17_translate/de-en',
                     evaluation_datasets: Optional[List[str]] = None,
                     reverse_translation=True,
                     shard_idx=0,
                     shard_count=1,
                     data_dir=None):
  """Load raw WMT datasets and normalize feature keys.

  Args:
    dataset_name: str: TFDS WMT dataset name.
    evaluation_datasets: Optional[List[[str]]: separate dataset names list for
      evaluation. e.g. for specifying the standard academic WMT14 test set.
    reverse_translation: bool: whether to reverse the translation direction.
      e.g. for 'de-en' this translates from english to german.
    shard_idx: int: for multihost training, index of this host.
    shard_count: int: for mulithost training, number of total hosts.
    data_dir: str: location of TFDS data directory.

  Returns:
    training tf.dataset, evaluation tf.dataset, and training features_info
    source and target language features are mapped to 'inputs' and 'targets'
    keys.
  """
  builder = tfds.builder(dataset_name, data_dir=data_dir)
  shard_spec = (f'[{int(100 * shard_idx / shard_count)}%'
                f':{int(100 * (shard_idx + 1) / shard_count)}%]')
  logging.info('Training on TFDS dataset %s with split %s', dataset_name,
               'train' + shard_spec)
  train_data = builder.as_dataset(
      split='train' + shard_spec, shuffle_files=True)

  eval_data = {}
  if evaluation_datasets is None:
    logging.info('Evaluating on TFDS dataset %s with split %s', dataset_name,
                 'validation' + shard_spec)
    # Using validation split of training dataset.
    dataset = builder.as_dataset(
        split='validation' + shard_spec, shuffle_files=False)
    eval_data[dataset_name + ':validation'] = dataset
  else:
    for name in evaluation_datasets:
      eval_dataset_name, *eval_split = name.split(':')
      if not eval_split:
        eval_split = 'validation'
        dataset_name = name + ':validation'
      else:
        eval_split = eval_split[0]
        dataset_name = name
      logging.info('Evaluating on TFDS dataset %s with split %s',
                   eval_dataset_name, eval_split + shard_spec)
      eval_builder = tfds.builder(eval_dataset_name, data_dir=data_dir)
      dataset = eval_builder.as_dataset(
          split=eval_split + shard_spec, shuffle_files=False)
      eval_data[dataset_name] = dataset

  # Train and eval set should have the same features
  features_info = builder.info

  # standardize on 'inputs' and 'targets' features.
  input_lang = features_info.supervised_keys[0]
  target_lang = features_info.supervised_keys[1]
  if reverse_translation:
    input_lang, target_lang = target_lang, input_lang

  def to_features_dict(x):
    return {'inputs': x[input_lang], 'targets': x[target_lang]}

  train_data = train_data.map(to_features_dict, num_parallel_calls=AUTOTUNE)
  standardized_eval_data = {
      name: dataset.map(to_features_dict, num_parallel_calls=AUTOTUNE)
      for name, dataset in eval_data.items()
  }

  return train_data, standardized_eval_data, features_info


# -----------------------------------------------------------------------------
# Tokenization.
# -----------------------------------------------------------------------------
def dump_chars_to_textfile(dataset,
                           maxchars=1e7,
                           data_keys=('inputs', 'targets')):
  """Write part of a TFDS sentence dataset to lines in a text file.

  Args:
    dataset: tf.dataset containing string-data.
    maxchars: int: approximate number of characters to save from dataset.
    data_keys: Tuple[str]: what keys in dataset to dump from.

  Returns:
    name of temp file with dataset bytes, exact number of characters dumped.
  """
  char_count = 0
  ds_iter = dataset.as_numpy_iterator()
  with tempfile.NamedTemporaryFile(
      delete=False, prefix='/tmp/ds_chars') as outfp:
    while char_count < maxchars:
      example = next(ds_iter)
      for k in data_keys:
        line = example[k] + b'\n'
        char_count += len(line)
        outfp.write(line)
  return outfp.name, char_count


def train_sentencepiece(dataset,
                        vocab_size,
                        maxchars=1e7,
                        character_coverage=1.0,
                        model_path='wmt_model.model',
                        model_type='unigram',
                        data_keys=('inputs', 'targets')):
  """Train SentencePiece tokenizer from subset of tf dataset.

  Args:
    dataset: tf.dataset
    vocab_size: int: size of vocab tokens to train.
    maxchars: int: number of characters to use for sentencepiece training.
    character_coverage: amount of characters covered by the model, good defaults
      are 0.9995 for languages with rich character set like Japanese or Chinese
      and 1.0 for other languages with small character set.
    model_path: str: path of model file to save vocab model to.
    model_type: str: type of sentencepiece vocab to train.
    data_keys: Tuple[str]: keys of dataset to use for training.

  Returns:
    path to the trained sentencepiece vocabulary model.
  """
  abs_model_path = os.path.abspath(os.path.expanduser(model_path))
  fname, _ = dump_chars_to_textfile(
      dataset, maxchars=maxchars, data_keys=data_keys)
  with tempfile.NamedTemporaryFile(
      delete=False, prefix='/tmp/sp_tmp') as model_fp:
    pass  # we just want a prefix'd tmp-filename
  argstr = ' '.join([
      f'--input={fname}', f'--vocab_size={vocab_size}',
      f'--character_coverage={character_coverage}',
      f'--model_prefix={model_fp.name}', f'--model_type={model_type}'
  ])
  SentencePieceTrainer.Train(argstr)
  # Only write to CNS if host id is 0 to prevent race conditions during
  # multihost training, otherwise wait until host 0 has written the file.
  if jax.host_id() == 0:
    # Use an intermediate filename that is renamed to the target name to address
    # create and fill delays.  Using finalization (CNS) as a indicator is not
    # portable.
    copy_rename_path = abs_model_path + '.rntmp'
    tf.io.gfile.copy(model_fp.name + '.model', copy_rename_path, overwrite=True)
    tf.io.gfile.rename(copy_rename_path, abs_model_path, overwrite=True)
    logging.info('copied %s to %s', model_fp.name + '.model', abs_model_path)
  else:
    while not tf.io.gfile.exists(abs_model_path):
      time.sleep(1)
    time.sleep(1)
  return abs_model_path


def load_sentencepiece_tokenizer(model_path,
                                 add_bos=False,
                                 add_eos=True,
                                 reverse=False):
  """Load a tf-text SentencePiece tokenizer from given model filepath."""
  with tf.io.gfile.GFile(model_path, 'rb') as model_fp:
    sp_model = model_fp.read()
  sp_tokenizer = tftxt.SentencepieceTokenizer(
      model=sp_model, add_bos=add_bos, add_eos=add_eos, reverse=reverse)
  logging.info('Loaded SentencePiece Tokenizer from file.')
  return sp_tokenizer


# -----------------------------------------------------------------------------
# Dynamic to static shape transforms.
# -----------------------------------------------------------------------------
def bin_and_batch(dataset,
                  n_devices,
                  batch_size=256,
                  bucket_length=32,
                  buckets=None,
                  drop_remainder=True):
  """Dynamic batching by length-bucketing.

  Sorts data into a small number of batch x length "buckets" that have roughly
  constant token count.

  Args:
    dataset: tf.data dataset
    n_devices: int: number of local devices
    batch_size: int: target batch size
    bucket_length: int: target length for target batch size
    buckets: List[Tuple[int, int]]: pairs of bucket-length, batch-size
      boundaries to define custom length-buckets.
    drop_remainder: bool: whether or not to drop the last odd-shaped batch
      produced by bucketing a finite input data stream.

  Returns:
    tf.data dataset with dynamically batched examples.
  """
  # Create heuristic buckets is none are specified.
  if buckets is None:
    logging.info('Heuristically bucketing based on shapes of examples.')
    bucket_boundaries = [
        bucket_length // 4, bucket_length // 2, bucket_length,
        bucket_length * 2, bucket_length * 4, bucket_length * 8,
        bucket_length * 16
    ]
    bucket_batch_sizes = [
        batch_size * 4, batch_size * 2, batch_size, batch_size // 2,
        batch_size // 4, batch_size // 8, batch_size // 16
    ]
    # TF.data's bucket_by_sequence_length pads to (bucket_boundary - 1):
    # we add 1 here to pad to the correct specified length.
    bucket_boundaries = [b + 1 for b in bucket_boundaries]
    # Make batch sizes divisible by n_devices.
    bucket_batch_sizes = [
        max(b // n_devices, 1) * n_devices for b in bucket_batch_sizes
    ]
    buckets = (bucket_boundaries, bucket_batch_sizes)

  logging.info('Bucketing with buckets %s.', str(buckets))

  def example_length(example):
    """The length function used by bucket_by_sequence_length to bucket."""
    return tf.maximum(
        tf.shape(example['inputs'])[0],
        tf.shape(example['targets'])[0])

  boundaries, batch_sizes = buckets
  # bucket_by_sequence_length expects a final dummy 1 batch_size.
  batch_sizes.append(1)
  dataset = dataset.apply(
      tf.data.experimental.bucket_by_sequence_length(
          example_length,
          boundaries,
          batch_sizes,
          pad_to_bucket_boundary=True,
          drop_remainder=drop_remainder))
  return dataset


def pack_dataset(dataset, length, keys=None):
  """Creates a 'packed' version of a dataset on-the-fly.

  Adapted from the mesh-tf implementation.

  This is meant to replace the irritation of having to create a separate
  "packed" version of a dataset to train efficiently on TPU.
  Each example in the output dataset represents several examples in the
  input dataset.
  For each key in the input dataset, two additional keys are created:
  <key>_segmentation: an int32 tensor identifying the parts
     representing the original example.
  <key>_position: an int32 tensor identifying the position within the original
     example.
  Example:
  Two input examples get combined to form an output example.
  The input examples are:
  {"inputs": [8, 7, 1, 0], "targets":[4, 1, 0]}
  {"inputs": [2, 3, 4, 1], "targets":[5, 6, 1]}
  The output example is:
  {
                 "inputs": [8, 7, 1, 2, 3, 4, 1, 0, 0, 0]
    "inputs_segmentation": [1, 1, 1, 2, 2, 2, 2, 0, 0, 0]
        "inputs_position": [0, 1, 2, 0, 1, 2, 3, 0, 0, 0]
                "targets": [4, 1, 5, 6, 1, 0, 0, 0, 0, 0]
   "targets_segmentation": [1, 1, 2, 2, 2, 0, 0, 0, 0, 0]
       "targets_position": [0, 1, 0, 1, 2, 0, 0, 0, 0, 0]
  }
  0 represents padding in both the inputs and the outputs.
  Sequences in the incoming examples are truncated to length "length", and the
  sequences in the output examples all have fixed (padded) length "length".

  Args:
    dataset: a tf.data.Dataset
    length: an integer, or a dict from feature-key to integer
    keys: a list of strings (e.g. ["inputs", "targets"])

  Returns:
    a tf.data.Dataset
  """
  shapes = tf.nest.map_structure(lambda spec: spec.shape, dataset.element_spec)
  if keys is None:
    keys = list(shapes.keys())
  for k in keys:
    if k not in shapes:
      raise ValueError('Key %s not found in dataset.  Available keys are %s' %
                       (k, shapes.keys()))
    if not shapes[k].is_compatible_with(tf.TensorShape([None])):
      raise ValueError('Tensors to be packed must be one-dimensional.')
  # make sure that the length dictionary contains all keys as well as the
  # keys suffixed by "_segmentation" and "_position"
  length_dict = {}
  for k in keys:
    for suffix in ['', '_segmentation', '_position']:
      length_dict[k + suffix] = length if isinstance(length, int) else length[k]
  length = length_dict

  # trim to length
  dataset = dataset.map(
      lambda x: {k: x[k][:length[k]] for k in keys},
      num_parallel_calls=AUTOTUNE)
  # Setting batch_size=length ensures that the concatenated sequences (if they
  # have length >=1) are sufficient to fill at least one packed example.
  batch_size = max(length.values())
  dataset = dataset.padded_batch(
      batch_size, padded_shapes={k: [-1] for k in keys})
  dataset = _pack_with_tf_ops(dataset, keys, length)

  # Set the Tensor shapes correctly since they get lost in the process.
  def my_fn(x):
    return {k: tf.reshape(v, [length[k]]) for k, v in x.items()}

  return dataset.map(my_fn, num_parallel_calls=AUTOTUNE)


def _pack_with_tf_ops(dataset, keys, length):
  """Helper-function for packing a dataset which has already been batched.

  Helper for pack_dataset()  Uses tf.while_loop.

  Args:
    dataset: a dataset containing padded batches of examples.
    keys: a list of strings
    length: an dict from feature-key to integer

  Returns:
    a dataset.
  """
  empty_example = {}
  for k in keys:
    empty_example[k] = tf.zeros([0], dtype=tf.int32)
    empty_example[k + '_position'] = tf.zeros([0], dtype=tf.int32)
  keys_etc = empty_example.keys()

  def write_packed_example(partial, outputs):
    new_partial = empty_example.copy()
    new_outputs = {}
    for k in keys_etc:
      new_outputs[k] = outputs[k].write(
          outputs[k].size(),
          tf.pad(partial[k], [[0, length[k] - tf.size(partial[k])]]))
    return new_partial, new_outputs

  def map_fn(x):
    """Internal function to flat_map over.

    Consumes a batch of input examples and produces a variable number of output
    examples.
    Args:
      x: a single example

    Returns:
      a tf.data.Dataset
    """
    partial = empty_example.copy()
    i = tf.zeros([], dtype=tf.int32)
    dynamic_batch_size = tf.shape(x[keys[0]])[0]
    outputs = {}
    for k in keys:
      outputs[k] = tf.TensorArray(
          tf.int32, size=0, dynamic_size=True, element_shape=[length[k]])
      outputs[k + '_position'] = tf.TensorArray(
          tf.int32, size=0, dynamic_size=True, element_shape=[length[k]])

    def cond_fn(i, partial, outputs):
      del partial, outputs
      return i < dynamic_batch_size

    def body_fn(i, partial, outputs):
      """Body function for while_loop.

      Args:
        i: integer scalar
        partial: dictionary of Tensor (partially-constructed example)
        outputs: dictionary of TensorArray

      Returns:
        A triple containing the new values of the inputs.
      """
      can_append = True
      one_example = {}
      for k in keys:
        val = tf.cast(x[k][i], tf.int32)
        val = val[:tf.reduce_sum(tf.cast(tf.not_equal(val, 0), tf.int32))]
        one_example[k] = val
      for k in keys:
        can_append = tf.logical_and(
            can_append,
            tf.less_equal(
                tf.size(partial[k]) + tf.size(one_example[k]), length[k]))

      def false_fn():
        return write_packed_example(partial, outputs)

      def true_fn():
        return partial, outputs

      partial, outputs = tf.cond(can_append, true_fn, false_fn)
      new_partial = {}
      for k in keys:
        new_seq = one_example[k][:length[k]]
        new_seq_len = tf.size(new_seq)
        new_partial[k] = tf.concat([partial[k], new_seq], 0)
        new_partial[k + '_position'] = tf.concat(
            [partial[k + '_position'],
             tf.range(new_seq_len, dtype=tf.int32)], 0)
      partial = new_partial
      return i + 1, partial, outputs

    i, partial, outputs = tf.while_loop(
        cond_fn,
        body_fn, (i, partial, outputs),
        shape_invariants=(
            tf.TensorShape([]),
            {k: tf.TensorShape([None]) for k in keys_etc},
            {k: tf.TensorShape(None) for k in keys_etc},
        ))
    partial, outputs = write_packed_example(partial, outputs)
    packed = {k: outputs[k].stack() for k in keys_etc}
    for k in keys:
      packed[k + '_segmentation'] = (
          tf.cumsum(
              tf.cast(tf.equal(packed[k + '_position'], 0), tf.int32), axis=1) *
          tf.cast(tf.not_equal(packed[k], 0), tf.int32))
    return packed

  dataset = dataset.map(map_fn, num_parallel_calls=AUTOTUNE)
  return dataset.unbatch()


# -----------------------------------------------------------------------------
# Main dataset prep routines.
# -----------------------------------------------------------------------------
def preprocess_wmt_data(dataset,
                        training,
                        n_devices,
                        dynamic_batching=False,
                        pack_examples=True,
                        shuffle_buffer_size=1024,
                        max_length=512,
                        batch_size=256,
                        bucket_length=32,
                        drop_remainder=True,
                        prefetch_size=AUTOTUNE,
                        seed: Optional[int] = None):
  """Shuffle and batch/pack the given dataset."""

  def length_filter(max_len):

    def filter_fn(x):
      source, target = x['inputs'], x['targets']
      l = tf.maximum(tf.shape(source)[0], tf.shape(target)[0])
      return tf.less(l, max_len + 1)

    return filter_fn

  if max_length > 0:
    dataset = dataset.filter(length_filter(max_length))

  if training:
    dataset = dataset.shuffle(shuffle_buffer_size, seed=seed)
    dataset = dataset.repeat()

  if pack_examples and dynamic_batching:
    raise ValueError(
        "Can't use both dynamic batching and packed-examples simultaneously.")

  if pack_examples:
    dataset = pack_dataset(dataset, max_length)
    dataset = dataset.batch(batch_size, drop_remainder=drop_remainder)
  elif dynamic_batching:
    dataset = bin_and_batch(
        dataset,
        n_devices,
        batch_size=batch_size,
        bucket_length=bucket_length,
        drop_remainder=drop_remainder)
  else:  # simple (static-shape) padded batching
    dataset = dataset.padded_batch(
        batch_size,
        padded_shapes={
            'inputs': max_length,
            'targets': max_length
        },
        padding_values={
            'inputs': 0,
            'targets': 0
        },
        drop_remainder=drop_remainder)

  if prefetch_size:
    dataset = dataset.prefetch(prefetch_size)

  return dataset


def get_wmt_datasets(
    n_devices,
    dataset_name='wmt17_translate/de-en',
    eval_dataset_list=None,
    reverse_translation=True,
    shard_idx=0,
    shard_count=1,
    data_dir=None,
    vocab_path=None,
    target_vocab_size=2**15,  # 32000
    max_corpus_chars=10**7,
    batch_size=256,
    bucket_length=32,
    dynamic_batching=False,
    pack_examples=True,
    max_length=256,
    max_eval_length=256,
    seed: Optional[int] = None):
  """Load and return dataset of batched examples for use during training."""
  if batch_size % n_devices:
    raise ValueError("Batch size %d isn't divided evenly by n_devices %d" %
                     (batch_size, n_devices))
  if vocab_path is None:
    vocab_path = os.path.expanduser('~/wmt_sentencepiece_model')

  train_data, eval_data_dict, _ = raw_wmt_datasets(
      dataset_name=dataset_name,
      evaluation_datasets=eval_dataset_list,
      reverse_translation=reverse_translation,
      shard_idx=shard_idx,
      shard_count=shard_count,
      data_dir=data_dir)

  try:
    sp_tokenizer = load_sentencepiece_tokenizer(vocab_path, add_eos=True)
  except tf.errors.NotFoundError:
    logging.info('SentencePiece vocab not found, building one from data.')
    vocab_dir = os.path.split(vocab_path)[0]
    tf.io.gfile.makedirs(vocab_dir)
    abs_vocab_path = train_sentencepiece(
        train_data,
        target_vocab_size,
        maxchars=max_corpus_chars,
        character_coverage=1.0,
        model_path=vocab_path,
        data_keys=('inputs', 'targets'))
    sp_tokenizer = load_sentencepiece_tokenizer(abs_vocab_path, add_eos=True)

  # Encode strings with sentencepiece tokenizer.
  def tokenize(data):
    return {
        'inputs': sp_tokenizer.tokenize(data['inputs']),
        'targets': sp_tokenizer.tokenize(data['targets'])
    }

  train_data = train_data.map(tokenize, num_parallel_calls=AUTOTUNE)
  eval_data_dict = {
      eval_dataset_name: eval_data.map(tokenize, num_parallel_calls=AUTOTUNE)
      for eval_dataset_name, eval_data in eval_data_dict.items()
  }

  train_batches = preprocess_wmt_data(
      train_data,
      training=True,
      dynamic_batching=dynamic_batching,
      pack_examples=pack_examples,
      n_devices=n_devices,
      batch_size=batch_size,
      bucket_length=bucket_length,
      max_length=max_length,
      seed=seed)

  preprocess_eval_data = functools.partial(
      preprocess_wmt_data,
      training=False,
      dynamic_batching=dynamic_batching,
      pack_examples=False,
      n_devices=n_devices,
      batch_size=batch_size,
      bucket_length=bucket_length,
      max_length=max_eval_length,
      seed=seed)

  eval_batches_dict = {
      eval_dataset_name: preprocess_eval_data(eval_data)
      for eval_dataset_name, eval_data in eval_data_dict.items()
  }

  # TODO(b/155933172): Would be better to refactor raw_wmt_datasets to get a
  # fixed train-eval split of training data. Currently it is changing on every
  # step.
  train_eval_batches = preprocess_wmt_data(
      train_data,
      training=False,
      dynamic_batching=dynamic_batching,
      pack_examples=False,
      n_devices=n_devices,
      batch_size=batch_size,
      bucket_length=bucket_length,
      max_length=max_length,
      seed=seed)

  preprocess_predict_data = functools.partial(
      preprocess_wmt_data,
      training=False,
      dynamic_batching=dynamic_batching,
      pack_examples=False,
      n_devices=n_devices,
      batch_size=batch_size,
      bucket_length=bucket_length,
      max_length=max_eval_length,
      drop_remainder=False,
      seed=seed)

  predict_batches_dict = {
      eval_dataset_name: preprocess_predict_data(eval_data)
      for eval_dataset_name, eval_data in eval_data_dict.items()
  }

  return train_batches, eval_batches_dict, train_eval_batches, predict_batches_dict, sp_tokenizer



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Flax implementation of ResNet V1 with optional quantization."""

import dataclasses
import functools
import typing
from typing import Any, Iterable, Optional, Tuple, Type
from aqt.jax_legacy.jax import flax_layers as aqt_flax_layers
from aqt.jax_legacy.jax import quant_config
from aqt.jax_legacy.jax.flax import struct as flax_struct
from flax import linen as nn
import jax.numpy as jnp

dataclass = flax_struct.dataclass if not typing.TYPE_CHECKING else dataclasses.dataclass
PRNGKey = Any
Shape = Iterable[int]
Dtype = Any
Array = Any


class ResidualBlock(nn.Module):
  """Bottleneck ResNet block."""

  @dataclass  # pylint: disable=missing-class-docstring
  class HParams:
    """Hyper params class for ResnetBlock."""
    # 'conv_proj' is only used in between different stages of residual blocks
    #   where the Tensor shape needs to be matched for the next stage.
    #   When it is not needed, it is set to None in hparams.
    conv_proj: Optional[aqt_flax_layers.ConvAqt.HParams]
    conv_se: aqt_flax_layers.ConvAqt.HParams  # unused in this model
    conv_1: aqt_flax_layers.ConvAqt.HParams
    conv_2: aqt_flax_layers.ConvAqt.HParams
    conv_3: aqt_flax_layers.ConvAqt.HParams
    act_function: str  # unused in this model
    shortcut_ch_shrink_method: str  # unused in this model
    shortcut_ch_expand_method: str  # unused in this model
    shortcut_spatial_method: str  # unused in this model

  hparams: HParams
  filters: int
  dynamic_context: quant_config.DynamicContext
  strides: Tuple[int, int]
  train: bool
  dtype: Type[Any]
  paxis_name: Optional[str] = 'batch'

  @nn.compact
  def __call__(
      self,
      inputs: jnp.ndarray,
  ) -> jnp.ndarray:
    """Applies a residual block consisting of Conv-Batch Norm-ReLU chains."""
    filters = self.filters
    strides = self.strides
    hparams = self.hparams
    dtype = self.dtype
    train = self.train
    dynamic_context = self.dynamic_context

    batch_norm = functools.partial(
        nn.BatchNorm,
        use_running_average=not train,
        momentum=0.9,
        epsilon=1e-5,
        dtype=dtype)

    conv = functools.partial(
        aqt_flax_layers.ConvAqt,
        use_bias=False,
        dtype=dtype,
        dynamic_context=dynamic_context,
        paxis_name=self.paxis_name,
        train=train)

    needs_projection = inputs.shape[-1] != filters * 4 or strides != (1, 1)
    assert needs_projection != (hparams.conv_proj is None)

    r1 = inputs
    if needs_projection:
      r1 = conv(
          features=filters * 4,
          kernel_size=(1, 1),
          strides=strides,
          name='proj_conv',
          hparams=hparams.conv_proj)(
              r1)
      r1 = batch_norm(name='proj_bn')(r1)

    def conv_block(
        inputs,
        n,
        features,
        kernel_size,
        strides,
        conv_hparams,
        r1=None,
    ):
      y = conv(
          features=features,
          kernel_size=kernel_size,
          name='conv' + n,
          strides=strides,
          hparams=conv_hparams)(
              inputs)
      scale_init = nn.initializers.zeros if n == '3' else nn.initializers.ones
      y = batch_norm(name='bn' + n, scale_init=scale_init)(y)
      if r1 is not None:
        y = r1 + y
      y = nn.relu(y)
      return y

    y = conv_block(
        inputs,
        n='1',
        features=filters,
        kernel_size=(1, 1),
        strides=(1, 1),
        conv_hparams=hparams.conv_1)

    y = conv_block(
        y,
        n='2',
        features=filters,
        kernel_size=(3, 3),
        strides=strides,
        conv_hparams=hparams.conv_2)

    y = conv_block(
        y,
        n='3',
        features=filters * 4,
        kernel_size=(1, 1),
        strides=(1, 1),
        conv_hparams=hparams.conv_3,
        r1=r1)

    return y


class ResNet(nn.Module):
  """ResNetV1 with optional quantization."""

  @dataclass
  class HParams:
    dense_layer: aqt_flax_layers.DenseAqt.HParams
    conv_init: aqt_flax_layers.ConvAqt.HParams
    residual_blocks: Tuple[ResidualBlock.HParams, ...]
    filter_multiplier: float
    act_function: str  # unused in this model
    se_ratio: float  # unused in this model
    init_group: int  # unused in this model

  num_classes: int
  hparams: HParams
  dynamic_context: quant_config.DynamicContext
  num_filters: int
  train: bool
  dtype: Type[Any]
  paxis_name: Optional[str] = 'batch'

  @nn.compact
  def __call__(
      self,
      inputs: jnp.ndarray,
  ) -> jnp.ndarray:
    """Applies ResNet model. Number of residual blocks inferred from hparams."""
    num_classes = self.num_classes
    hparams = self.hparams
    num_filters = self.num_filters
    dtype = self.dtype

    x = aqt_flax_layers.ConvAqt(
        features=num_filters,
        kernel_size=(7, 7),
        strides=(2, 2),
        padding=[(3, 3), (3, 3)],
        use_bias=False,
        dtype=dtype,
        name='init_conv',
        train=self.train,
        dynamic_context=self.dynamic_context,
        paxis_name=self.paxis_name,
        hparams=hparams.conv_init,
    )(
        inputs)
    x = nn.BatchNorm(
        use_running_average=not self.train,
        momentum=0.9,
        epsilon=1e-5,
        dtype=dtype,
        name='init_bn')(
            x)
    x = nn.relu(x)
    x = nn.max_pool(x, (3, 3), strides=(2, 2), padding='SAME')
    filter_multiplier = hparams.filter_multiplier
    for i, block_hparams in enumerate(hparams.residual_blocks):
      proj = block_hparams.conv_proj
      # For projection layers (unless it is the first layer), strides = (2, 2)
      if i > 0 and proj is not None:
        filter_multiplier *= 2
        strides = (2, 2)
      else:
        strides = (1, 1)
      x = ResidualBlock(
          filters=int(num_filters * filter_multiplier),
          hparams=block_hparams,
          dynamic_context=self.dynamic_context,
          strides=strides,
          train=self.train,
          dtype=dtype)(
              x)

    x = jnp.mean(x, axis=(1, 2))

    x = aqt_flax_layers.DenseAqt(
        features=num_classes,
        dtype=dtype,
        train=self.train,
        dynamic_context=self.dynamic_context,
        paxis_name=self.paxis_name,
        hparams=hparams.dense_layer,
    )(x, padding_mask=None)

    x = jnp.asarray(x, dtype)
    # The output of ViT does not have log_softmax.
    # To make resnet50 teacher has the same type of outputs as ViT,
    # comment out the following line
    # output = nn.log_softmax(x)
    return x


def create_resnet(hparams, dtype, train, **kwargs):
  return ResNet(
      num_classes=1000,
      dtype=dtype,
      hparams=hparams,
      dynamic_context=quant_config.DynamicContext(
          update_bounds=False, quantize_weights=True),
      num_filters=64,
      train=train,
      **kwargs)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Create a sweep over half_shift flag under different weight precisions.

During the sweep, only weights are quantized.
The weight precisions of conv_init and the final dense layer are set to 8.
"""

import copy
from aqt.jax_legacy.jax.imagenet.configs import base_config
import ml_collections


def get_config():
  """Returns sweep configuration (see module docstring)."""
  sweep_config = ml_collections.ConfigDict()
  base_config_dict = base_config.get_config(
      imagenet_type=base_config.ImagenetType.RESNET50,
      quant_target=base_config.QuantTarget.WEIGHTS_ONLY)
  configs = []

  for half_shift in [False, True]:
    for prec in [1, 2, 3, 4]:
      config = copy.deepcopy(base_config_dict)
      config.weight_prec = prec
      config.model_hparams.conv_init.weight_prec = 8
      config.model_hparams.dense_layer.weight_prec = 8
      config.half_shift = half_shift
      config.metadata.hyper_str = f"resnet50_w_only_halfshift_sweep_halfshift_{half_shift}_prec_{prec}"
      configs.append(config)

  sweep_config.configs = configs
  return sweep_config
